import tkinter as tk
import tkinter.filedialog as tkfd
import tkinter.messagebox as tkmb
from tkinter import ttk

import os
from subprocess import Popen

from sol.inputs import osc
from sol.database import database, clip

from sol.setup_gui import SetupGui
from sol.config import GlobalConfig
C = GlobalConfig()

from .sub_wins import RenameWin, MoveWin, TagWin, AddTagWin, Treeview


class LibraryOrgGui:
    def __init__(self, root, parent, standalone=False):

        # gui holders
        self.parent = parent  # the big gui
        self.child_wins = {}
        self.add_clip_gui = None
        self.setup_gui = None

        # sol stuff
        self.backend = self.parent.magi
        self.db = self.backend.db
        self.osc_client = None
        self.clip_storage_dict = None
        self.last_selected_clip = None

        # class data
        self.standalone = standalone
        self.folders = {}
        self.all_folder_names = []
        self.all_tags = []

        self.delayed_actions = []
        self.total_deleted_count = 0

        self.sub_process = None

        if C.MODEL_SELECT == 'MPV':
            self.last_selected_clip = self.backend.clip_storage.current_clips[0]

            def clip_selector(clip, layer):
                self.backend.select_clip(clip, layer)
                self.perform_delayed_actions(clip)

            def clip_clearer(*args):
                self.backend.clear_clip(0)
                self.last_selected_clip = None

        elif C.XTERNAL_PLAYER_SELECT == 'MEMEPV' and os.path.exists(C.MEMEPV_SCRIPT_PATH):
            Popen(['node', C.MEMEPV_SCRIPT_PATH, '1', '6999'])
            self.osc_client = osc.OscClient(port=6999)

            def clip_selector(clip, layer=0):
                self.osc_client.build_n_send('/0/load', clip.f_name)
                self.perform_delayed_actions(clip)

            def clip_clearer(*args):
                self.osc_client.build_n_send('/0/clear', 1)
                self.last_selected_clip = None

        else:
            def clip_selector(clip, layer=0):
                if '{}' in C.EXTERNAL_PLAYER_COMMAND:
                    command = C.EXTERNAL_PLAYER_COMMAND.format(clip.f_name)
                else:
                    command = C.EXTERNAL_PLAYER_COMMAND + ' "{}"'.format(clip.f_name)
                try:
                    self.sub_process = Popen(command)
                except:
                    if C.DEBUG:
                        print('your external command doesnt work gg')
                    return
                self.perform_delayed_actions(clip)

            def clip_clearer(*args):
                if self.sub_process is not None:
                    self.sub_process.kill()
                    self.sub_process = None

        self.select_clip = clip_selector
        self.clear_clip = clip_clearer

        ###############
        # SETUP THE GUI

        # main window
        self.root = root
        self.root.title('lib org')

        if standalone:
            self.root.call = self.parent.root.call

        self.mainframe = ttk.Frame(root, padding='0')
        self.mainframe.pack(fill=tk.BOTH, expand=tk.Y)

        self.parent.root.call('wm', 'attributes', '.', '-topmost', '0')
        self.root.protocol("WM_DELETE_WINDOW", self.parent.exit_lib_org_gui)
        self.root.lift()
        self.root.focus_force()

        # treeview
        self.tree_frame = ttk.Frame(self.mainframe)
        self.tree_frame.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.tree = Treeview(self.tree_frame, self.root)
        self.tree.tree.bind('<Double-1>', self.activate_clip)
        self.tree.tree.bind('<Return>', self.activate_clip)
        self.tree.tree.bind('<Double-3>', self.clear_clip)

        # menubar
        self.menubar = tk.Menu(self.root)
        self.filemenu = tk.Menu(self.menubar, tearoff=0)

        self.filemenu.add_command(label="save", command=self.save_prompt, accelerator='Ctrl-S')
        self.filemenu.add_command(label="save as", command=self.save_as_prompt, accelerator='Ctrl-Shift-S')
        self.root.bind("<Control-s>", lambda e: self.save_prompt())
        self.root.bind("<Control-Shift-KeyPress-S>", lambda e: self.save_as_prompt())

        self.filemenu.add_command(label="load", command=self.load_prompt, accelerator='Ctrl-O')
        self.root.bind("<Control-o>", lambda e: self.load_prompt())

        if standalone:
            self.filemenu.add_command(label="edit options", command=self.edit_config)

        self.menubar.add_cascade(label='file', menu=self.filemenu)
        self.menubar.add_command(label="import wizard", command=self.create_clip_gui)
        self.menubar.add_command(label="refresh (f5)", command=self.init_tree)
        self.root.bind("<F5>", self.init_tree)

        self.root.config(menu=self.menubar)
        # self.root.geometry('1000x500+50+100')  # temp.....

        # bottombar
        bot_but_pad = '0 5 0 5'
        self.bottom_bar = ttk.Frame(self.root)
        self.bottom_bar.pack(side=tk.BOTTOM, anchor=tk.S, fill=tk.X, expand=True)

        self.delete_but = ttk.Button(self.bottom_bar, text='(DEL)ete', padding=bot_but_pad, command=self.delete_prompt)
        self.root.bind("<Delete>", self.delete_prompt)

        self.rename_but = ttk.Button(self.bottom_bar, text='(R)ename', padding=bot_but_pad, command=self.rename_prompt)
        self.root.bind("r", self.rename_prompt)
        self.root.bind("<F2>", self.rename_prompt)

        self.tag_but = ttk.Button(self.bottom_bar, text='(T)ag', padding=bot_but_pad, command=self.tag_prompt)
        self.root.bind("t", self.tag_prompt)

        self.move_but = ttk.Button(self.bottom_bar, text='(M)ove', padding=bot_but_pad, command=self.move_prompt)
        self.root.bind("m", self.move_prompt)

        for but in [self.delete_but, self.rename_but, self.tag_but, self.move_but, ]:
            but.pack(side=tk.LEFT)

        # LAST STEPS...

        if standalone:
            self.load()
        else:
            self.gen_all_folder_names()
            self.clip_storage_dict = self.db.file_ops.save_clip_storage(self.backend.clip_storage)

        self.init_tree()

    def init_tree(self, *args):
        self.be_safe()
        files = self.db.hierarchical_listing
        for folder in self.folders.values():
            if self.tree.tree.exists(folder):
                self.tree.tree.delete(folder)
        if files is None:
            return
        self.folders = {}
        cur_folder = ''
        for i in range(len(files)):
            node = files[i]
            if node[0] == 'folder':
                top_folder = ''
                if node[2] in self.folders:
                    top_folder = self.folders[node[2]]
                    self.tree.tree.item(top_folder, open=True)
                self.folders[node[1]] = cur_folder = self.tree.tree.insert(top_folder, 'end', text=node[1], open=False, values=['', '', 'folder'])

            else:
                fname = files[i][2]
                tags = self.db.clips[fname].str_tags()
                self.tree.tree.insert(cur_folder, 'end', text=files[i][1], values=[tags, fname, 'clip'])

    ###############
    # other windows

    def create_clip_gui(self):
        if self.add_clip_gui is None:
            self.add_clip_gui = ImportWizardGui(tk.Toplevel(takefocus=True), self)
        else:
            self.add_clip_gui.root.lift()
            self.add_clip_gui.root.focus_force()

    def edit_config(self):
        if self.setup_gui is not None:
            self.setup_gui.root_frame.focus_force()
        else:
            self.setup_gui = SetupGui(parent=self)

    def toggle_on_top(self, *args):
        # for config options..
        pass

    ####################
    # NEATER SUBWIN FUNS

    def get_clip_and_i(self):
        clip_fname, row_id = self.tree.get_selected_clip()
        if clip_fname is not None and len(clip_fname) > 0:
            if clip_fname in self.db.clips:
                return self.db.clips[clip_fname], row_id
        return None, None

    def callback_gen(self, clip, row_id, what_kind):
        stor_clip = clip
        i = row_id

        if what_kind in ['move', 'rename']:
            def callback(new_path, new_name=None):
                hold_vals = self.tree.tree.item(i)['values']
                old_path = hold_vals[1]
                hold_vals[1] = new_path
                self.tree.tree.item(i, values=hold_vals)
                if new_name is not None:
                    self.tree.tree.item(i, text=new_name)

                def maybe_do_later():
                    if os.path.exists(old_path):
                        def rename_later():
                            try:
                                # actually rename the file?
                                os.rename(old_path, new_path)
                                # change clip's fname/name
                                self.db.move_clip(stor_clip, new_path, new_name)
                            except:
                                pass
                        self.parent.root.after(1000, rename_later)

                if self.last_selected_clip == stor_clip:
                    self.delayed_actions += [(stor_clip, maybe_do_later)]
                else:
                    maybe_do_later()

        else:  # tag
            def callback(new_tags):
                self.db.tagdb.update_clip_tags(new_tags, stor_clip)
                self.db.tagdb.refresh()

                new_tags = stor_clip.str_tags()
                hold_vals = self.tree.tree.item(i)['values']
                hold_vals[0] = new_tags

                # change row in the treeview
                self.tree.tree.item(i, values=hold_vals)

        return callback

    #################
    # ACTION PROMPTS

    def move_prompt(self, *args):
        clip, row_id = self.get_clip_and_i()
        if clip is None:
            return
        callback = self.callback_gen(clip, row_id, 'move')
        MoveWin(self, clip, callback)

    def rename_prompt(self, *args):
        clip, row_id = self.get_clip_and_i()
        if clip is None:
            return
        callback = self.callback_gen(clip, row_id, 'rename')
        RenameWin(self, clip, callback)

    def tag_prompt(self, *args):

        def gen_tag_list_callback(cliplist, itemlist):
            clipz = cliplist
            iz = itemlist

            def gend_fun(tag_list):
                for i, clp in enumerate(clipz):
                    self.db.tagdb.update_clip_tags(tag_list, clp)
                    new_tags = clp.str_tags()
                    hold_vals = self.tree.tree.item(iz[i])['values']
                    hold_vals[0] = new_tags
                    self.tree.tree.item(iz[i], values=hold_vals)
                self.db.tagdb.refresh()

            return gend_fun

        selected_rows = self.tree.tree.selection()
        selected_ids = []
        selected_clips = []
        for r in selected_rows:
            selected_item = self.tree.tree.item(r)['values']
            if selected_item[2] == 'clip':
                clip_fname = selected_item[1]
                if (len(clip_fname) > 0) and selected_item[1] in self.db.clips:
                    selected_clips.append(self.db.clips[clip_fname])
                    selected_ids.append(r)

        if len(selected_ids) == 0:
            return
        elif len(selected_ids) == 1:
            callback = self.callback_gen(selected_clips[0], selected_ids[0], 'tag')
            TagWin(self, selected_clips[0], callback)
        else:
            AddTagWin(self, selected_clips, gen_tag_list_callback(selected_clips, selected_ids))

    ##########
    # DELETING

    def delete_prompt(self, event=None):

        selected_rows = self.tree.tree.selection()
        for sr in selected_rows:
            check_break = self.delete_selected_row(sr)
            if not check_break:
                return

    def delete_selected_row(self, row_id):
        try:
            selected_item = self.tree.tree.item(row_id)
        except:  # if we deleted the folder earlier row_id wont exist anymore
            return True
        if selected_item['values'][2] == 'folder':
            find_folder = selected_item['text']
            # get the full folder path
            check_folders = [fp[0].split(os.sep)[-1] for fp in self.all_folder_names]
            extra_string = find_folder
            found_i = -1
            if find_folder in check_folders:
                found_i = check_folders.index(find_folder)
                full_path = self.all_folder_names[found_i][1]
                extra_string += "\t({})".format(full_path)
            yes_no = tkmb.askyesno('delete', 'are you sure you want to delete this folder\n{}?'.format(extra_string),
                                   icon='warning', default='no')
            if yes_no:
                fnames_to_delet = self.tree.delet_all_children(row_id)
                if found_i >= 0:
                    del self.all_folder_names[found_i]
                for fn in fnames_to_delet:
                    self.delete_clip_fname(fn)
                    self.total_deleted_count += 1
                return True
            else:
                return False
        else:
            clip = self.tree.delet_selected_clip(row_id)
            if clip is not None:
                clip_fname = clip['values'][1]
                self.delete_clip_fname(clip_fname)
                self.total_deleted_count += 1
        return True

    def delete_clip_fname(self, fname):
        if fname in self.db.clips:
            self.db.remove_clip(self.db.clips[fname])

    #########
    # MOVING

    def gen_all_folder_names(self):
        self.all_folder_names = []
        hl = self.db.hierarchical_listing
        for i in range(len(hl) - 1):
            if hl[i][0] == 'folder':
                if hl[i + 1][0] == 'clip':
                    self.add_a_folder_name(os.path.split(hl[i + 1][2])[0], hl[i][1])
        # if C.DEBUG: print(self.all_folder_names)

    def add_a_folder_name(self, f_path, f_name=None):
        if f_name is None:
            f_name = os.path.split(f_path)[1]

        just_fnames = [os.path.split(f)[1] for f, _ in self.all_folder_names]

        if f_name in just_fnames:
            # find all matches
            indices = [i for i, f in enumerate(just_fnames) if f == f_name]

            # find unique .. path
            # first put all subfolders into a list of lists
            split_paths = []

            def gen_flist(f):
                f_list = os.path.normpath(f).split(os.sep)
                f_list = [p for p in f_list if p not in C.IGNORED_DIRS]
                f_list.reverse()
                return f_list

            new_path_list = gen_flist(f_path)
            split_paths.append(new_path_list)

            min_len = len(new_path_list)

            for i in indices:
                f_list = gen_flist(self.all_folder_names[i][1])
                split_paths.append(f_list)
                min_len = min(min_len, len(f_list))

            # now find first index that is unique
            found_i = -1
            for i in range(min_len):
                check_subpaths = [fl[i] for fl in split_paths]
                if len(check_subpaths) == len(set(check_subpaths)):
                    found_i = i
                    break

            # now update the foldernames
            for i in range(len(split_paths)):
                if found_i > -1:
                    paths_to_join = split_paths[i][:found_i + 1][::-1]
                    paths_to_join.insert(0, '..')
                else:
                    paths_to_join = split_paths[i][::-1]

                new_fname = os.sep.join(paths_to_join)

                if i == 0:
                    # new folder
                    self.all_folder_names.append([new_fname, f_path])
                else:
                    # old ones
                    self.all_folder_names[indices[i - 1]][0] = new_fname
        else:
            self.all_folder_names.append([f_name, f_path])

    ##########
    # TAGGING

    def gen_all_tags(self):
        self.all_tags = self.db.all_tags[:]

    def add_a_tag(self, tag):
        self.all_tags.append(tag)
        self.all_tags.sort()

    ############
    # CLIP STUFF

    def activate_clip(self, event=None):
        clip_fname, _ = self.tree.get_selected_clip(event)
        if clip_fname is None or len(clip_fname) < 1:
            return
        # print(clip_fname)
        if clip_fname in self.db.clips:
            self.select_clip(self.db.clips[clip_fname], 0)

    ###########
    # SAVE/LOAD

    def save_prompt(self, filename=None):
        if filename is None:
            filename = self.db.file_ops.last_save
        if filename is None:
            self.save_as_prompt()
        else:
            self.save(filename)

    def save_as_prompt(self):
        ask_fun = tkfd.asksaveasfilename
        filename = ask_fun(parent=self.root, title='Save as..', initialdir=C.SAVEDATA_DIR)
        if filename:
            self.save(filename)

    def load_prompt(self, *args):
        ask_fun = tkfd.askopenfilename
        filename = ask_fun(parent=self.root, title='Load', initialdir=C.SAVEDATA_DIR)
        if filename:
            if self.load(filename):
                self.refresh()

    def be_safe(self):
        for clip_action_pair in self.delayed_actions:
            if clip_action_pair[0] == self.last_selected_clip:
                self.clear_clip()
            clip_action_pair[1]()
        self.delayed_actions = []

    def save(self, filename=None):
        fio = self.db.file_ops
        if filename is None:
            filename = fio.last_save
        if filename is None:
            return
        self.be_safe()
        try:
            root = fio.create_save('magi')
            root.append(self.clip_storage_dict)
            root.append(fio.save_database(self.db))
            save_data = fio.pretty_print(root)
            with open(filename, 'wb') as f:
                f.write(save_data)
            fio.update_last_save(filename)
            if C.DEBUG:
                print('successfully saved', filename)
        except Exception as e:
            if C.DEBUG:
                print(e)
            print('failed to save', filename)

    def load(self, filename=None):
        self.be_safe()
        try:
            fio = self.db.file_ops
            if filename is None:
                filename = fio.last_save

            parsed_xml = fio.create_load(filename)
            # load the database
            self.db.clear()
            fio.load_database(parsed_xml.find('database'), self.db)
            # save clip storage
            # this needs to be done separately in the case of standalone liborg
            self.clip_storage_dict = parsed_xml.find('clip_storage')
            fio.update_last_save(filename)
            if C.DEBUG:
                print('successfully loaded', fio.last_save)
            # update all folders & tags
            self.gen_all_folder_names()
            self.gen_all_tags()
            self.total_deleted_count = 0
            return True
        except Exception as e:
            if C.DEBUG:
                print(e)
            print('failed to load')
            return False

    #####################
    # MISC HELPER METHODS

    def refresh(self):
        self.tree.clear()
        self.init_tree()

    def close(self, *args):
        self.parent.root.call('wm', 'attributes', '.', '-topmost', str(int(self.parent.on_top_toggle.get())))
        self.quit()

    def quit(self, *args):
        self.be_safe()
        if self.osc_client is not None:
            self.osc_client.build_n_send('/0/quit', True)
        if self.standalone:
            yes_no = True
            if self.total_deleted_count > 10:
                yes_no = tkmb.askyesno('overwrite save',
                                       'you\'ve deleted a lot of clips ({} to be exact)\nare you sure you want to overwrite your savefile?'.format(self.total_deleted_count),
                                       icon='warning', default='no')
            if yes_no:
                self.save()
            else:
                self.save_as_prompt()
        self.root.destroy()

    def perform_delayed_actions(self, clip):
        self.last_selected_clip = clip
        new_delayed_actions = []
        for clip_action_pair in self.delayed_actions:
            if clip_action_pair[0] != clip:
                clip_action_pair[1]()
            else:
                new_delayed_actions += [clip_action_pair]
        self.delayed_actions = new_delayed_actions


class ImportWizardGui:
    def __init__(self, top_frame, parent):
        self.parent = parent
        self.backend = parent.backend
        self.db = self.backend.db

        self.all_folder_names = self.parent.all_folder_names
        self.add_a_folder_name = self.parent.add_a_folder_name
        self.all_tags = self.parent.all_tags
        self.add_a_tag = self.parent.add_a_tag

        # so i can keep track of which clip is selected
        self.clip_queue = []
        self.fname_to_clip = {}
        self.child_wins = self.parent.child_wins

        self.root = top_frame
        self.root.title('import wizard')

        self.tree = Treeview(self.root, self.root, select_mode='browse', enabled_cols=[0, 2])
        self.tree.tree.bind('<Double-1>', self.activate_clip)
        self.tree.tree.bind('<Return>', self.activate_clip)
        self.tree.tree.bind('<Double-3>', self.parent.clear_clip)

        self.menubar = tk.Menu(self.root)
        self.menubar.add_command(label="add folder", command=self.add_folder_prompt)
        self.menubar.add_command(label="import to library", command=self.do_import)
        self.menubar.add_command(label="clear", command=self.clear_all)
        self.menubar.add_command(label="quit", command=self.quit)
        self.root.bind_all("<Control-q>", self.quit)
        self.root.config(menu=self.menubar)
        self.root.protocol("WM_DELETE_WINDOW", self.quit)

        bot_but_pad = '0 5 0 5'
        self.bottom_bar = ttk.Frame(self.root)
        self.bottom_bar.pack(side=tk.BOTTOM, anchor=tk.S, fill=tk.X, expand=True)

        self.root.bind("<Delete>", self.delete_clip)
        self.delete_but = ttk.Button(self.bottom_bar, text='(DEL)ete', padding=bot_but_pad, command=self.delete_clip)

        self.root.bind("r", self.rename_clip)
        self.root.bind("<F2>", self.rename_clip)
        self.rename_but = ttk.Button(self.bottom_bar, text='(R)ename', padding=bot_but_pad, command=self.rename_clip)

        self.root.bind("t", self.tag_clip)
        self.tag_but = ttk.Button(self.bottom_bar, text='(T)ag', padding=bot_but_pad, command=self.tag_clip)

        self.root.bind("m", self.move_clip)
        self.move_but = ttk.Button(self.bottom_bar, text='(M)ove', padding=bot_but_pad, command=self.move_clip)

        for but in [self.delete_but, self.rename_but, self.tag_but, self.move_but, ]:
            but.pack(side=tk.LEFT)

        self.root.lift()
        self.root.focus_force()

    ##############
    # ADDING CLIPS

    def add_folder_prompt(self):
        ask_fun = tkfd.askdirectory
        foldername = ask_fun(parent=self.root, title='add folder', mustexist=True)
        foldername = os.sep.join(foldername.split('/'))
        self.add_folder(foldername)

    def add_folder(self, folder):
        new_clips = []
        if folder:
            for item in os.listdir(folder):
                full_path = os.path.join(folder, item)
                if not os.path.isdir(full_path):
                    if any([full_path.lower().endswith(f_ext) for f_ext in C.SUPPORTED_FILETYPES]):
                        new_clips += [full_path]
        for c in new_clips:
            new_clip = clip.Clip(c, '0')
            self.db.init_a_clip(new_clip)
            new_clip.params['play_direction'] = 'f'

            self.clip_queue += [new_clip]
            self.add_clip_to_list(new_clip)

    def do_import(self):
        self.parent.be_safe()
        for c in self.clip_queue:
            self.db.add_clip(c)
        self.db.searcher.refresh()
        self.parent.refresh()
        self.clear_all()

    def add_clip_to_list(self, clip):
        self.fname_to_clip[clip.f_name] = clip
        self.tree.tree.insert('', 'end', text=clip.name,
                              values=[clip.str_tags(), clip.f_name, 'clip'])

    ################
    # MORE CLIPS FUN

    def activate_clip(self, event=None):
        clip_fname, _ = self.tree.get_selected_clip(event)
        if clip_fname is None or len(clip_fname) < 1:
            return
        # print(clip_fname)
        if clip_fname in self.fname_to_clip:
            actual_clip = self.fname_to_clip[clip_fname]
            self.parent.select_clip(actual_clip, 0)

    ##########
    # DELETING

    def delete_clip(self, event=None):
        clip = self.tree.delet_selected_clip()
        if clip is not None:
            clip_fname = clip['values'][1]
            if clip_fname in self.fname_to_clip:
                to_del = self.fname_to_clip[clip_fname]
                del self.clip_queue[self.clip_queue.index(to_del)]
                del self.fname_to_clip[clip_fname]

    #####################
    # NEATER SUBWIN FUNS

    def get_clip_and_i(self):
        clip_fname, row_id = self.tree.get_selected_clip()
        if clip_fname not in self.fname_to_clip:
            return None, None
        actual_clip = self.fname_to_clip[clip_fname]
        return actual_clip, row_id

    def callback_gen(self, clip, row_id, what_kind):
            stor_clip = clip
            i = row_id

            if what_kind in ['move', 'rename']:
                def callback(new_path, new_name=None):
                    hold_vals = self.tree.tree.item(i)['values']
                    old_path = hold_vals[1]
                    hold_vals[1] = new_path
                    self.tree.tree.item(i, values=hold_vals)
                    if new_name is not None:
                        self.tree.tree.item(i, text=new_name)

                    # update fname_to_clip..
                    if old_path in self.fname_to_clip:
                        del self.fname_to_clip[old_path]
                    self.fname_to_clip[new_path] = clip

                    def maybe_do_later():
                        if os.path.exists(old_path):
                            def rename_later():
                                try:
                                    # actually rename the file?
                                    os.rename(old_path, new_path)
                                    # change clip's fname/name
                                    stor_clip.f_name = new_path
                                    if new_name is not None:
                                        stor_clip.name = new_name
                                except:
                                    pass
                            self.parent.root.after(1000, rename_later)

                    if self.parent.last_selected_clip == stor_clip:
                        self.parent.delayed_actions += [(stor_clip, maybe_do_later)]
                    else:
                        maybe_do_later()

            else:  # tag
                def callback(new_tags):
                    self.db.tagdb.update_clip_tags(new_tags, stor_clip)
                    self.db.tagdb.refresh()

            return callback

    ################
    # ACTION PROMPTS

    def rename_clip(self, *args):
        clip, row_id = self.get_clip_and_i()
        if clip is None:
            return
        callback = self.callback_gen(clip, row_id, 'rename')
        RenameWin(self, clip, callback)

    def move_clip(self, *args):
        clip, row_id = self.get_clip_and_i()
        if clip is None:
            return
        callback = self.callback_gen(clip, row_id, 'move')
        MoveWin(self, clip, callback)

    def tag_clip(self, *args):
        clip, row_id = self.get_clip_and_i()
        if clip is None:
            return
        callback = self.callback_gen(clip, row_id, 'tag')
        TagWin(self, clip, callback)

    ##############
    # MISC HELPERS

    def clear_all(self):
        self.clip_queue = []
        self.fname_to_clip = {}
        self.tree.clear()

    def quit(self, *args):
        self.root.destroy()
        self.parent.add_clip_gui = None
        self.parent.be_safe()


if __name__ == '__main__':

    class FakeParent():
        def __init__(self, root):
            self.root = root
            self.magi = self
            self.db = database.Database()
            self.on_top_toggle = tk.BooleanVar()
            self.on_top_toggle.set(False)
            self.child = None

            def do_nothing(*args):
                pass
            self.select_clip = do_nothing

        def exit_lib_org_gui(self, *args):
            if self.child is not None:
                self.child.close()
            self.root.destroy()

    rootwin = tk.Tk()
    rootwin.title('lib_org')
    rootwin.withdraw()

    fp = FakeParent(rootwin)
    saliborg = LibraryOrgGui(tk.Toplevel(takefocus=True), fp, standalone=True)
    fp.child = saliborg
    # saliborg.create_clip_gui()
    # saliborg.add_clip_gui.add_folder('C:\\VJ\\zzz_incoming_clips')
    rootwin.mainloop()



import os

import tkinter as tk
import tkinter.filedialog as tkfd
import tkinter.messagebox as tkmb

from tkinter import ttk

from sol.magi import Magi
from sol.gui.lib_org import lib_org

# from sol.gui.tk_gui import clip_control
from sol.gui.tk_gui import clip_control
from sol.gui.tk_gui import clip_collections
from sol.gui.tk_gui import midi_config
from sol.setup_gui import SetupGui

from sol.mt_gui import MTGui

from sol.config import GlobalConfig
C = GlobalConfig()


class MainGui:
    """
    test gui made with tkinter and re-using many parts
    of past sol versions
    """

    def __init__(self, root):

        # tk
        self.root = root
        # theming
        C.themer.setup(C.CURRENT_THEME, root)
        self.mainframe = ttk.Frame(root, padding='0')
        self.cc_frame = ttk.Frame(self.mainframe, padding='0')
        self.cc_frames = []
        self.clip_col_frame = ttk.Frame(self.mainframe, padding='0')

        # sol
        self.magi = Magi()
        self.magi.gui = self
        if C.MTGUI_ENABLED:
            self.mt_gui = MTGui(self.magi, ip=C.MTGUI_IP_ADDR)
        else:
            self.mt_gui = None
        self.setup_gui = None
        self.midi_config_gui = None
        self.clip_controls = []
        self.clip_org = None
        self.lib_org = None

        for i in range(C.NO_LAYERS):
            new_frame = ttk.Frame(self.cc_frame, padding='0')
            self.clip_controls.append(
                clip_control.ClipControl(new_frame, self.magi, i))
            self.cc_frames.append(new_frame)
        self.clip_conts = clip_collections.CollectionsHolder(
            self.root, self.clip_col_frame, self.magi)

        # pack it
        self.mainframe.pack()
        self.cc_frame.pack(side=tk.TOP, fill=tk.X, expand=True)
        for frame in self.cc_frames:
            frame.pack(side=tk.LEFT)
        self.clip_col_frame.pack(side=tk.TOP, fill=tk.BOTH)

        # menu
        self.setup_menubar()
        # quit behavior
        self.root.protocol("WM_DELETE_WINDOW", self.quit)

        self.root.resizable(0, 0)

    def start(self):
        try:
            self.magi.start()
        except Exception as e:
            print(e)
            tkmb.showerror('fatal error', e)
            self.quit()

    def new_save(self):
        self.magi.reset()
        self.refresh_after_load()

    def save(self, filename=None):
        if filename is None:
            filename = self.magi.db.file_ops.last_save
        if filename is None:
            self.save_as()
        else:
            self.magi.save_to_file(filename)

    def save_as(self):
        ask_fun = tkfd.asksaveasfilename
        filename = ask_fun(parent=self.root, title='Save as..',
                           initialdir=C.SAVEDATA_DIR)
        if filename:
            self.save(filename)

    def load(self):
        ask_fun = tkfd.askopenfilename
        filename = ask_fun(parent=self.root, title='Load',
                           initialdir=C.SAVEDATA_DIR)
        if filename:
            if self.magi.load(filename):
                self.refresh_after_load()
            else:
                self.new_save()

    def load_resolume(self):
        ask_fun = tkfd.askopenfilename
        filename = ask_fun(
            parent=self.root, title='Open Resolume Composition', initialdir=C.RESOLUME_SAVE_DIR)
        if filename:
            self.magi.load_resolume_comp(filename)
            # self.gen_thumbs() # don't do this by default actually
            self.refresh_after_load()

    def export_resolume(self):
        ask_fun = tkfd.askopenfilename
        filename = ask_fun(
            parent=self.root, title='Starting Resolume Composition', initialdir=C.RESOLUME_SAVE_DIR)
        ask_fun = tkfd.asksaveasfilename
        filename_out = ask_fun(
            parent=self.root, title='Save as..', initialdir=C.RESOLUME_SAVE_DIR)
        if filename:
            self.magi.export_resolume_comp(filename, filename_out)

    def load_isadora(self):
        ask_fun = tkfd.askdirectory
        foldername = ask_fun(
            parent=self.root, title='Add Folder', mustexist=True)
        if foldername:
            self.magi.load_isadora_comp(foldername)
            # self.gen_thumbs()
            self.refresh_after_load()

    def gen_thumbs(self):
        self.magi.gen_thumbs(n_frames=C.NO_FRAMES)
        self.refresh_after_load()

    def refresh_after_load(self):
        self.clip_conts.clip_storage = self.magi.clip_storage
        self.clip_conts.refresh_after_load()
        for i in range(C.NO_LAYERS):
            self.clip_controls[i].update_clip(
                self.magi.clip_storage.current_clips[i])
        # print([0],self.magi.clip_storage.current_clips[1])

    def configure_midi(self):
        if self.midi_config_gui is not None:
            self.midi_config_gui.root.focus_force()
        else:
            self.midi_config_gui = midi_config.MidiConfig(tk.Toplevel(), self)

    def edit_config(self):
        if self.setup_gui is not None:
            self.setup_gui.root_frame.focus_force()
        else:
            self.setup_gui = SetupGui(parent=self)
        # if always on top disable it temporarily (mayb useful for other things
        # too...)

    def toggle_on_top(self, *args):
        new_val = self.on_top_toggle.get()
        new_val = str(int(new_val))
        self.root.call('wm', 'attributes', '.', '-topmost', new_val)

    def exit_org_gui(self):
        # resize the 1st clip control back to normal
        self.clip_controls[0].resize(330)
        for i in range(1, len(self.cc_frames)):
            self.cc_frames[i].pack(side=tk.LEFT)

        # repack bottom half
        self.clip_conts.col_frame.pack_forget()
        self.clip_conts.search_frame.pack(side=tk.LEFT, fill=tk.Y, anchor='e')
        self.clip_conts.col_frame.pack(side=tk.LEFT, fill=tk.X, expand=True)

        self.cur_view.set('full')

    def enter_clip_org_gui(self):
        # only want 1 clip control =)
        for i in range(1, len(self.cc_frames)):
            self.cc_frames[i].pack_forget()

        # resize the 1 clip control
        self.clip_controls[0].resize(4 * C.THUMB_W, True)
        # we want different library browser sry
        self.clip_conts.search_frame.pack_forget()
        self.clip_org = clip_collections.ClipOrg(tk.Toplevel(), self)

    def exit_clip_org_gui(self, *args):
        # close clip_org if it isn't closed yet
        if self.clip_org is not None:
            self.clip_org.close()
            self.clip_org = None

        # reset
        self.exit_org_gui()

    def enter_lib_org_gui(self):
        # close clip_org if it isn't closed yet
        if self.lib_org is not None:
            self.lib_org.close()
            self.lib_org = None
        # only want 1 clip control =)
        for i in range(1, len(self.cc_frames)):
            self.cc_frames[i].pack_forget()

        # resize the 1 clip control
        self.clip_controls[0].resize(4 * C.THUMB_W, True)
        # we want different library browser sry
        self.clip_conts.search_frame.pack_forget()
        self.lib_org = lib_org.LibraryOrgGui(tk.Toplevel(), self)

    def exit_lib_org_gui(self, *args):
        # close lib_org if it isn't closed yet
        if self.lib_org is not None:
            self.lib_org.close()
            self.lib_org = None

        # reset
        self.exit_org_gui()

    def change_views(self, *args):
        new_view = self.cur_view.get()
        enter_actions = {
            'clip_org': self.enter_clip_org_gui,
            'lib_org': self.enter_lib_org_gui
        }

        exit_actions = {
            'clip_org': self.exit_clip_org_gui,
            'lib_org': self.exit_lib_org_gui
        }

        # exit last
        if self.last_view in exit_actions:
            exit_actions[self.last_view]()

        # enter new
        if new_view in enter_actions:
            enter_actions[new_view]()

        self.last_view = new_view

    def setup_menubar(self):
        self.menubar = tk.Menu(self.root)
        self.filemenu = tk.Menu(self.menubar, tearoff=0)  # file menu
        # new
        self.filemenu.add_command(label="new", command=self.new_save)
                                  # accelerator='Ctrl-N') # how to add this text
        # save
        self.filemenu.add_command(label="save", command=self.save)
        # save as
        self.filemenu.add_command(label="save as", command=self.save_as)
        # load
        self.filemenu.add_command(label="load", command=self.load)
        if C.MODEL_SELECT == 'RESOLUME':
            self.filemenu.add_command(
                label="load resolume comp", command=self.load_resolume)
            self.filemenu.add_command(
                label="export resolume comp", command=self.export_resolume)

        elif C.MODEL_SELECT == 'ISADORA':
            self.filemenu.add_command(
                label="add folder (isadora)", command=self.load_isadora)
        # create thumbnails
        self.filemenu.add_command(
            label="generate thumbs", command=self.gen_thumbs)
        # launch midi configurator
        self.filemenu.add_command(
            label="config midi", command=self.configure_midi)
        # launch the setup utility
        self.filemenu.add_command(
            label="edit options", command=self.edit_config)
        # quit
        self.filemenu.add_command(label="quit", command=self.quit)

        self.viewmenu = tk.Menu(self.menubar, tearoff=0)  # view menu
        self.on_top_toggle = tk.BooleanVar()
        self.on_top_toggle.trace('w', self.toggle_on_top)
        self.on_top_toggle.set(C.ALWAYS_ON_TOP)

        self.cur_view = tk.StringVar()
        self.last_view = ''
        self.cur_view.trace('w', self.change_views)

        # toggle always on top behavior
        self.viewmenu.add_checkbutton(
            label="always on top", onvalue=True, offvalue=False, variable=self.on_top_toggle)

        # switch between the different views
        # full, clip_org, performance
        self.viewmenu.add_separator()
        self.viewmenu.add_radiobutton(
            label='full view', value='full', variable=self.cur_view)
        self.viewmenu.add_radiobutton(
            label='clip org view', value='clip_org', variable=self.cur_view)
        self.viewmenu.add_radiobutton(
            label='lib org view', value='lib_org', variable=self.cur_view)

        # self.viewmenu.add_radiobutton(label='performace view',value='perf',variable=self.cur_view)
        self.cur_view.set('full')
        # pack it
        self.menubar.add_cascade(label='file', menu=self.filemenu)
        self.menubar.add_cascade(label='view', menu=self.viewmenu)
        self.root.config(menu=self.menubar)

    def quit(self):
        self.save()
        self.magi.stop()
        if self.lib_org is not None:
            self.lib_org.close()
        self.root.destroy()

    # funs required by magi

    def update_clip(self, layer, clip):
        # if clip is none.. clear
        self.clip_controls[layer].update_clip(clip)
        if self.mt_gui is not None:
            self.mt_gui.update_clip(layer, clip)

    def update_clip_params(self, layer, clip, param):
        # dispatch things according to param
        try:
            self.clip_controls[layer].update_clip_params(clip, param)
            if self.mt_gui is not None:
                self.mt_gui.update_clip_params(layer, clip, param)
        except:
            pass

    def update_cur_pos(self, layer, pos):
        # pass along the current position
        self.clip_controls[layer].update_cur_pos(pos)
        if self.mt_gui is not None:
            self.mt_gui.update_cur_pos(layer, pos)

    def update_search(self):
        # display search results
        self.clip_conts.library_browse.last_search()
        if self.mt_gui is not None:
            self.mt_gui.update_search()  # doesn't do anything yet

    def update_cols(self, what, ij=None):
        # what - select, add, remove, swap
        # ij - what index (swap returns a tuple..)
        what_to_do = {
            'add': self.clip_conts.add_collection_frame,
            'select': self.clip_conts.highlight_col,
            'swap': self.clip_conts.swap,
            'remove': self.clip_conts.remove_collection_frame
        }

        if what in what_to_do:
            if ij is None:
                what_to_do[what]()
            else:
                what_to_do[what](ij)

        if self.mt_gui is not None:
            self.mt_gui.update_cols(what, ij)

    def update_clip_names(self):
        for i in range(C.NO_LAYERS):
            if self.magi.clip_storage.current_clips[i] is not None:
                self.clip_controls[i].change_name(
                    self.magi.clip_storage.current_clips[i].name)
        for c_i in range(len(self.magi.clip_storage.clip_cols)):
            active_col = self.magi.clip_storage.clip_cols[c_i]
            for i in range(len(active_col)):
                if active_col[i] is None:
                    pass
                else:
                    self.clip_conts.containers[c_i]\
                        .clip_conts[i].change_text(active_col[i].name)

        if self.mt_gui is not None:
            self.mt_gui.update_clip_names()  # also doesn't do anything

    def update_gui_directly(self, layer, what, n=-1):
        self.clip_controls[layer].update_directly(what, n)


def main():
    root = tk.Tk()
    root.title('sol')

    if not os.path.exists(C.config_savefile):
        tkmb.showinfo(
            '', 'no config found!\n you need to run initial setup first')
        SetupGui(root)

    else:
        testgui = MainGui(root)

        testgui.refresh_after_load()

        testgui.start()

    root.mainloop()


if __name__ == '__main__':

    main()



import os


class Clip:
    """
    used to store "video clips"
    with associated information
    most importantly, filepath and how to activate them
            but also params and tags

    example params (inspired by previous sol version)
            cue_points - point in time to jump to
            loop_points - collection of pairs of points between which can loop
             -> loop_type - default (d) or bounce (b)
            loop_selection - which loop points are chosen
            loop_on - boolean
            playback_speed - self explanatory
            play_direction - forward (f), backward (b), pause (p), random (r)
            control_speed - multiplication factor for controlling clip
    """

    def __init__(self, filepath, activation_string, name=None,
                 thumbnail_path=None, params={}, tags=[]):

        self.f_name = filepath
        self.t_names = thumbnail_path

        self.command = activation_string
        self.params = params
        self.tags = []

        if name is None:
            self.name = os.path.split(os.path.splitext(filepath)[0])[1]
        else:
            self.name = name

        for tag in tags:
            self.add_tag(tag)

    def add_tag(self, tag):
        if tag not in self.tags:
            self.tags.append(tag)

    def str_tags(self):
        temp_tags = self.tags[:]
        temp_tags.sort()
        return ','.join(temp_tags)

    def remove_tag(self, tag):
        if tag in self.tags:
            self.tags.remove(tag)

    def __lt__(self, other):
        # needed for sort
        return self.f_name < other

    def __gt__(self, other):
        # needed for sort
        return self.f_name > other

    def __str__(self):
        return self.name + \
            "\n\t@ " + self.f_name


class ClipCollection:
    """
    a collection of n many clips
    """

    def __init__(self, n=8, name='new_col'):
        self.name = name
        self.clips = [None] * n

    def __getitem__(self, i):
        return self.clips[i]

    def __setitem__(self, i, clip):
        self.safe_extend(i)
        self.clips[i] = clip

    def __len__(self):
        return len(self.clips)

    def index(self, what):
        return self.clips.index(what)

    def safe_extend(self, i):
        i_diff = i - self.__len__()
        if i_diff >= 0:
            add_none = [None] * (4 * ((i_diff // 4) + 1))
            self.clips.extend(add_none)



import os

class IsadoraLoader:
    def __init__(self,total_clipz=0):
        self.clips = []
        self.last_clip_access = total_clipz

    def add_folder(self,folder):
        for item in os.listdir(folder):
            full_path = os.path.join(folder,item)
            if not os.path.isdir(full_path):
                self.clips += [full_path]


    def get_clips(self):
        tor = []
        for i, clip in enumerate(self.clips[self.last_clip_access:]):
            tor += [(clip, str(1 + i + self.last_clip_access))]
        self.last_clip_access = len(self.clips)
        return tor

if __name__ == '__main__':
    test = IsadoraLoader()
    test.add_folder("C:\\VJ\\__clips__\\artsy\\dxv")
    print(test.get_clips())
    test.add_folder("C:\\VJ\\__clips__\\ppl\\dxv")
    print(test.get_clips())



import tkinter.font as tkFont
from tkinter import ttk

def hex_to_rgb(hex_val):
    hex_val = hex_val.lstrip('#')
    if len(hex_val) < 6:
        hex_val = hex_val * 2
    return tuple(int(hex_val[i:i + 2], 16) for i in range(0, 6, 2))


def rgb_to_hex(rgb_val):
    return '#%02x%02x%02x' % rgb_val


def linerp_helper(c1, c2, d):
    return tuple(int((c2[i] - c1[i]) * d + c1[i]) for i in range(3))


def linerp_colors(two_colors, n):
    # input is 2 hex vals, outputs n many hex vals between them
    if (n <= 2):
        return two_colors
    from_col, to_col = hex_to_rgb(two_colors[0]), hex_to_rgb(two_colors[1])
    return [rgb_to_hex(linerp_helper(from_col, to_col, (x / (n - 1)))) for x in range(n)]


def setup(sel_theme, tk_root):
    styling = ttk.Style()
    styling.theme_use(sel_theme.default_style)

    if sel_theme.default_font is not None:
        default_font = tkFont.nametofont("TkDefaultFont")
        default_font.configure(**sel_theme.default_font)
        tk_root.option_add("*Font", default_font)
        styling.configure('.', font=default_font)
        styling.configure('Treeview.Heading', font=default_font)
        # (menus are system font unfortunately)

    if sel_theme.default_colors is not None:
        tk_root.tk_setPalette(**sel_theme.default_colors)
        styling.configure('.', **sel_theme.default_colors)

    for so in sel_theme.ttk_style_opts_to_config:
        styling.configure(so[0], **so[1])

    for som in sel_theme.ttk_style_opts_to_map:
        styling.map(som[0], **som[1])

    # stylistic choices for everything
    styling.layout("Treeview", [
        ('Treeview.treearea', {'sticky': 'nswe'})
    ])



import tkinter as tk
from tkinter import ttk
from .tkdnd import dnd_start
import tkinter.simpledialog as tksimpledialog
import tkinter.font as tkFont

from PIL import ImageTk, Image

import os
import bisect
from itertools import accumulate

from sol.config import GlobalConfig
C = GlobalConfig()

EMPTY_CLIP = os.path.join(os.path.dirname(__file__), 'sample_clip.png')


class ClipContainer:
    # gui element that holds a single clip
    def __init__(self, parent, index, selectfun=None, clip=None, first=False):
        self.index = index  # index in parent collection
        self.selectfun = selectfun
        self.last_clip = None

        self.parent = parent
        self.backend = parent.backend
        self.root = parent.root
        self.frame = ttk.Frame(self.parent.frame, padding='2 0 2 4')
        self.grid = self.frame.grid

        self.imgs = []
        self.default_img = self.current_img = ImageTk.PhotoImage(Image.open(EMPTY_CLIP))
        self.current_img_i = 0
        self.hovered = False
        if first:
            self.label = tk.Label(self.frame, image=self.current_img, text='test', compound='top',
                                  width=C.THUMB_W, borderwidth='2', name='coln_clip_{}'.format(index))  # width of clip preview
        else:
            self.label = tk.Label(self.frame, image=self.current_img, text='test', compound='top', width=C.THUMB_W, borderwidth='2')  # width of clip preview            
        self.label.image = self.current_img
        self.label.pack()
        self.label.bind('<Double-1>', self.activate_l)
        self.label.bind('<Double-3>', self.activate_r)
        self.label.bind("<Enter>", self.start_hover)
        self.label.bind("<Leave>", self.end_hover)

        self.frame.dnd_accept = self.dnd_accept

        self.clip = None
        self.change_clip(clip)

    # things to do with the clip
    def activate(self, *args, layer=-1):
        if self.clip is None or self.selectfun is None:
            return
        self.selectfun(self.clip, layer)

    def activate_l(self, *args):
        self.activate(*args, layer=0)

    def activate_r(self, *args):
        self.activate(*args, layer=1)

    def change_text(self, new_text):
        new_text = str(new_text)
        text_meas = []
        for c in new_text:
            if c in C.FONT_WIDTHS:
                text_meas.append(C.FONT_WIDTHS[c])
            else:
                try:
                    new_meas = self.parent.font_checker.measure(c)
                    C.FONT_WIDTHS[c] = new_meas
                    text_meas.append(new_meas)
                except:
                    text_meas.append(C.FONT_AVG_WIDTH)
        cumm_text_meas = list(accumulate(text_meas))
        if cumm_text_meas[-1] > C.THUMB_W:
            to_i = bisect.bisect_left(cumm_text_meas, C.THUMB_W - 5 * C.FONT_WIDTHS['.'])
            new_text = new_text[:to_i].strip() + ".."
        self.label.config(text=new_text)

    def change_clip(self, clip):
        if clip is None:
            self.remove_clip()
            return
        self.last_clip = self.clip
        self.clip = clip
        # update the clip in parent
        if self.parent.index >= 0:
            self.backend.clip_storage.clip_cols[self.parent.index][self.index] = clip

        if self.clip.t_names is None:
            self.change_img_from_img(self.default_img)
        else:
            self.setup_imgs()
        self.toggle_dnd()

    def remove_clip(self, *args):
        self.clip = None
        self.imgs = []
        self.hovered = False
        self.toggle_dnd()
        if self.parent.index >= 0:
            self.backend.clip_storage.clip_cols[self.parent.index][self.index] = None

    # thumbnail funs
    def change_img_from_file(self, new_img):
        self.current_img = ImageTk.PhotoImage(Image.open(new_img))
        self.label.config(image=self.current_img)  # necessary to do twice to preserve image across garbage collection
        self.label.image = self.current_img

    def change_img_from_img(self, new_img):
        self.current_img = new_img
        self.label.config(image=self.current_img)
        self.label.image = self.current_img

    def next_img(self):
        if len(self.imgs) == 0:
            return
        self.current_img_i = (self.current_img_i + 1) % len(self.imgs)
        self.change_img_from_img(self.imgs[self.current_img_i])

    def regen_imgs(self, *args):
        if self.clip is None:
            return
        new_pix = self.parent.backend.thumb_maker.gen_thumbnail(self.clip.f_name, n_frames=C.NO_FRAMES)
        if len(new_pix) == 0:
            return
        self.clip.t_names = new_pix
        self.setup_imgs()

    def hover_animate(self, *args):
        if len(self.imgs) == 0:
            return
        if not self.hovered:
            self.current_img_i = 0
            self.change_img_from_img(self.imgs[self.current_img_i])
            return
        self.next_img()
        self.root.after(C.REFRESH_INTERVAL, self.hover_animate)

    def start_hover(self, *args):
        self.hovered = True
        self.hover_animate()

    def end_hover(self, *args):
        self.hovered = False

    def setup_imgs(self):
        f_names = [os.path.join(C.SCROT_DIR, t_name) for t_name in self.clip.t_names
                   if os.path.exists(os.path.join(C.SCROT_DIR, t_name))]
        self.imgs = [ImageTk.PhotoImage(Image.open(f_name)) for f_name in f_names]
        if len(self.imgs) == 0:
            return
        self.current_img_i = 0
        self.change_img_from_img(self.imgs[self.current_img_i])

    # drag-n-drop stuff here
    def press(self, event):
        if dnd_start(DragClip(self.clip, self), event):
            pass

    def dnd_accept(self, source, event):
        return self

    def dnd_enter(self, source, event):
        pass

    def dnd_motion(self, source, event):
        pass

    def dnd_leave(self, source, event):
        pass

    def dnd_commit(self, source, event):
        if source.clip is None:
            return
        self.change_clip(source.clip)
        self.dnd_leave(source, event)

    def dnd_end(self, target, event):
        pass

    def toggle_dnd(self):
        if not self.clip:
            self.change_text('no_clip')
            self.change_img_from_img(self.default_img)
            self.label.unbind('<ButtonPress-1>')  # this disables selecting clip
            self.label.unbind('<ButtonPress-3>')  # this disables drag around
            self.label.unbind('<ButtonRelease-1>')
        else:
            self.change_text(self.clip.name)
            # change image, if clip_name is empty / our clip is none, set the img to default img -.-
            self.label.bind("<ButtonPress-3>", self.press, add="+")  # now we can drag it around
            self.label.bind('<ButtonPress-2>', self.remove_clip)  # middle click 2 remove clip


class DragClip:
    """
    used for dragging clips wherever
    """

    def __init__(self, clip, source=None):
        self.clip = clip
        self.source = source

    def dnd_end(self, target, event):
        if self.source and isinstance(target, type(self.source)):
            self.source.change_clip(target.last_clip)


class ContainerCollection:
    # gui element that holds multiple clips
    def __init__(self, root, parent_frame, clipcol, select_cmd, backend, first=False):
        self.clip_conts = []
        self.clip_collection = clipcol
        self.frame = ttk.Frame(parent_frame, padding=0)
        self.root = root
        self.backend = backend
        self.font_checker = tkFont.Font()

        if '.' not in C.FONT_WIDTHS:
            C.FONT_WIDTHS['.'] = self.font_checker.measure('.')

        cc_len = len(self.clip_collection)
        for i in range(C.NO_Q):
            start_clip = None
            if i < cc_len:
                start_clip = self.clip_collection[i]
            self.clip_conts.append(ClipContainer(self, i, select_cmd, start_clip, first))

        n_buts = C.NO_Q
        n_rows = 1
        if n_buts > 4:
            n_rows = n_buts // 4
            if n_buts % 4 != 0:
                n_rows += 1

        for r in range(n_rows):
            for c in range(4):
                i = r * 4 + c
                self.clip_conts[i].grid(row=r, column=c)

        self.frame.grid(row=0, column=0)
        self.frame.tkraise()

    def update_clip_col(self, clip_col):
        self.clip_collection = clip_col
        for i in range(C.NO_Q):
            self.clip_conts[i].change_clip(clip_col[i])

    def clear(self):
        for clip_cont in self.clip_conts:
            clip_cont.remove_clip()

    @property
    def index(self):
        if self.clip_collection in self.backend.clip_storage.clip_cols:
            return self.backend.clip_storage.clip_cols.index(self.clip_collection)
        return -1


class CollectionsHolder:
    # gui element that holds multiple containercollections
    def __init__(self, root, parent_frame, backend):
        self.backend = backend
        self.clip_storage = backend.clip_storage
        self.select_cmd = backend.select_clip

        self.root = root
        s = ttk.Style()
        defaultbg = s.configure('.')['background']
        self.frame = parent_frame
        self.col_frame = ttk.Frame(self.frame)
        self.collections_frame = ttk.Frame(self.col_frame, padding='0')

        self.collections_bottom_frame = ttk.Frame(self.col_frame, padding='0 1 0 1')
        self.collection_label_canvas = tk.Canvas(self.collections_bottom_frame, borderwidth=0,
                                                 background=defaultbg, width=(3 * C.THUMB_W + 20), height=C.FONT_HEIGHT + 2)
        self.collections_labels_frame = ttk.Frame(self.collection_label_canvas, padding='0')
        self.collection_label_canvas.create_window((0, 0), window=self.collections_labels_frame, anchor="nw")

        s.configure('colbar.Horizontal.TScrollbar', arrowsize=C.FONT_HEIGHT + 2)

        self.xsb = ttk.Scrollbar(self.collections_bottom_frame, orient='horizontal', name='coln_select',
                                 style='colbar.Horizontal.TScrollbar', command=self.collection_label_canvas.xview)

        self.collection_label_canvas.configure(xscrollcommand=self.xsb.set)

        self.collections_labels_frame.bind("<Configure>", lambda event: self.canvas_scroll_update())

        self.search_frame = ttk.Frame(self.frame)

        self.library_browse = LibraryBrowser(self.backend, self.search_frame)

        self.containers = []
        self.container_labels = []

        self.add_col_but = ttk.Button(self.collections_bottom_frame, text='+', command=self.add_collection, width=1)
        self.add_col_but.pack(side=tk.RIGHT)

        self.collections_frame.pack(side=tk.TOP)
        self.collections_bottom_frame.pack(side=tk.TOP, fill=tk.X, expand=False)
        self.collection_label_canvas.pack(side=tk.LEFT, fill=tk.Y, expand=False, anchor='s')
        self.xsb.pack(side=tk.LEFT, fill=tk.X, expand=True)
        self.search_frame.pack(side=tk.LEFT, fill=tk.BOTH, anchor='e', expand=True)
        self.col_frame.pack(side=tk.LEFT, fill=tk.X, expand=False)

        self.xsb.bind('<Button-3>', self.xsb_hax)

    # interact with backend
    # collection maintenance
    def add_collection(self):
        self.clip_storage.add_collection()
        self.highlight_col(len(self.clip_storage.clip_cols) - 1)

    def remove_collection(self, index):
        if len(self.containers) <= 1:
            self.containers[0].clear()
            return
        self.clip_storage.remove_collection(index)

    def select_collection(self, index):
        if index != self.clip_storage.cur_clip_col:
            self.clip_storage.select_collection(index)

    # swaps
    def go_left(self, *args):
        self.clip_storage.go_left()

    def go_right(self, *args):
        self.clip_storage.go_right()

    def swap_left(self, *args):
        self.clip_storage.swap_left()
        self.clip_storage.go_left()

    def swap_right(self, *args):
        self.clip_storage.swap_right()
        self.clip_storage.go_right()

    # gui updating
    def refresh_after_load(self):
        # clip collections
        for c in self.containers:
            c.frame.destroy()
        self.containers = []
        for l in self.container_labels:
            l.destroy()
        self.container_labels = []

        for collection in self.clip_storage.clip_cols:
            self.add_collection_frame(collection)

        self.highlight_col()
        # search
        self.library_browse.tree_reset()

    def swap(self, ij):
        if len(ij) != 2:
            return
        # labels
        text_i = self.container_labels[ij[0]].cget('text')
        text_j = self.container_labels[ij[1]].cget('text')
        self.container_labels[ij[0]].config(text=text_j)
        self.container_labels[ij[1]].config(text=text_i)
        # do the swap
        self.containers[ij[0]], self.containers[ij[1]] = self.containers[ij[1]], self.containers[ij[0]]
        self.highlight_col()

    def highlight_col(self, index=-1):
        if index < 0:
            index = self.clip_storage.cur_clip_col
        self.containers[index].frame.tkraise()
        for cl in self.container_labels:
            cl.configure(relief=tk.RIDGE, state='')
        self.container_labels[index].configure(relief=tk.SUNKEN, state='disabled')
        # scroll into view
        left_x, wid_width = self.container_labels[index].winfo_x(), self.container_labels[index].winfo_width()
        total_width = self.collection_label_canvas.bbox("all")[2]
        # check if need to scroll
        xsb_get = self.xsb.get()
        above, below = xsb_get[1] * total_width, xsb_get[0] * total_width
        if (left_x + wid_width >= above) or (left_x < below):
            scroll_to = (left_x - wid_width) / total_width
            self.collection_label_canvas.xview('moveto', scroll_to)

    # gui helpers
    def add_collection_frame(self, collection=None):
        if collection is None:
            collection = self.clip_storage.clip_cols[-1]
        new_cont = ContainerCollection(self.root, self.collections_frame, collection,
                                       self.select_cmd, self.backend, len(self.containers)==0)
        self.containers.append(new_cont)
        self.add_collection_label(collection)

    def add_collection_label(self, collection):
        newlabel = ttk.Label(self.collections_labels_frame, text=collection.name,
                             justify=tk.CENTER, padding='4 2 4 1', borderwidth='4',
                             width=-4, anchor='s', style='fakebut.TLabel')
        newlabel.bind('<ButtonPress-1>', lambda *args: self.select_collection(self.container_labels.index(newlabel)))
        newlabel.bind("<Double-1>", lambda *args: self.change_name_dialog(self.container_labels.index(newlabel)))
        newlabel.bind('<ButtonPress-2>', lambda *args: self.remove_collection(self.container_labels.index(newlabel)))
        newlabel.bind("<Double-3>", lambda *args: self.remove_collection(self.container_labels.index(newlabel)))
        newlabel.bind("<MouseWheel>", self.canvas_scroll_mouse)
        newlabel.bind("<Button-4>", self.canvas_scroll_mouse)
        newlabel.bind("<Button-5>", self.canvas_scroll_mouse)
        newlabel.pack(side=tk.LEFT, anchor='s', fill=tk.Y, expand=True)
        self.container_labels.append(newlabel)

    def remove_collection_frame(self, index):
        self.containers[index].frame.destroy()
        del self.containers[index]
        self.container_labels[index].destroy()
        del self.container_labels[index]
        self.highlight_col()

    def change_name_dialog(self, index):
        new_name = tksimpledialog.askstring("rename collection", '',
                                            initialvalue=self.containers[index].clip_collection.name)
        if new_name and new_name != self.containers[index].clip_collection.name:
            # change name
            self.containers[index].clip_collection.name = new_name
            self.container_labels[index].configure(text=new_name)

    # scrolling
    def canvas_scroll_update(self):
        self.collection_label_canvas.configure(scrollregion=self.collection_label_canvas.bbox("all"))

    def canvas_scroll_mouse(self, event):
        # first check if we want to scroll
        total_width = float(self.collection_label_canvas.config()['width'][-1])
        label_frame_width = self.collection_label_canvas.bbox("all")[2]
        if (total_width < label_frame_width):
            self.collection_label_canvas.xview('scroll', -1 * int(event.delta // 120), 'units')

    def xsb_hax(self, event):
        click_what = self.xsb.identify(event.x, event.y)
        if click_what == 'leftarrow':
            self.swap_left()
        elif click_what == 'rightarrow':
            self.swap_right()


# this stuff belongs in another class imo, nevermind it uses clipcontainers : )
class LibraryBrowser:
    def __init__(self, backend, parent_frame):
        self.backend = backend
        # db (has searcher & search fun)
        self.db = self.backend.db
        self.folders = {}

        self.parent_frame = parent_frame
        self.frame = ttk.Frame(self.parent_frame)  # , relief='groove')
        self.browsers = ttk.Notebook(self.frame)

        self.search_frame = ttk.Frame(self.parent_frame)
        self.browse_frame = ttk.Frame(self.parent_frame)
        self.tag_frame = ttk.Frame(self.parent_frame)

        # search tree setup
        self.search_query = tk.StringVar()
        self.search_field = ttk.Entry(self.search_frame, textvariable=self.search_query, width=5)
        self.search_query.trace('w', self.search)
        self.search_inner_frame = ttk.Frame(self.search_frame)
        self.search_tree = ttk.Treeview(self.search_inner_frame, selectmode='extended', show='tree', height=11)
        self.search_tree.bind('<ButtonPress>', self.make_drag_clip, add="+")
        self.search_tree.bind('<Double-1>', lambda e: self.activate_clip_to_layer(e, 0))
        self.search_tree.bind('<Double-3>', lambda e: self.activate_clip_to_layer(e, 1))

        self.search_field.pack(side=tk.TOP, anchor=tk.N, fill=tk.X, pady=2)
        self.search_tree.pack(side=tk.LEFT, anchor=tk.N, fill=tk.BOTH, expand=tk.Y)
        self.ysb = ttk.Scrollbar(self.search_frame, orient='vertical', command=self.search_tree.yview)
        self.search_tree.configure(yscrollcommand=self.ysb.set)
        self.ysb.pack(side=tk.RIGHT, anchor=tk.N, fill=tk.Y)
        self.search_inner_frame.pack(side=tk.TOP, anchor=tk.N, fill=tk.BOTH, expand=True)

        # browse tree setup
        self.browse_tree = ttk.Treeview(self.browse_frame, selectmode='extended', show='tree', height=13)
        self.browse_tree.bind('<ButtonPress>', self.make_drag_clip, add="+")
        self.browse_tree.bind('<Double-1>', lambda e: self.activate_clip_to_layer(e, 0))
        self.browse_tree.bind('<Double-3>', lambda e: self.activate_clip_to_layer(e, 1))

        self.browse_tree.pack(side=tk.LEFT, anchor=tk.N, fill=tk.BOTH, expand=tk.Y)
        self.ysbb = ttk.Scrollbar(self.browse_frame, orient='vertical', command=self.browse_tree.yview)
        self.browse_tree.configure(yscrollcommand=self.ysbb.set)
        self.ysbb.pack(side=tk.RIGHT, anchor=tk.N, fill=tk.Y)

        # tags tree setup
        self.tag_query = tk.StringVar()
        self.tag_field = ttk.Entry(self.tag_frame, textvariable=self.tag_query, width=5)
        self.tag_query.trace('w', self.tag_search)
        self.tag_inner_frame = ttk.Frame(self.tag_frame)
        self.tag_tree = ttk.Treeview(self.tag_inner_frame, selectmode='extended', show='tree', height=11)
        self.tag_tree.bind('<ButtonPress>', self.make_drag_clip, add="+")
        self.tag_tree.bind('<Double-1>', lambda e: self.activate_clip_to_layer(e, 0))
        self.tag_tree.bind('<Double-3>', lambda e: self.activate_clip_to_layer(e, 1))

        self.tag_field.pack(side=tk.TOP, anchor=tk.N, fill=tk.X, pady=2)
        self.tag_tree.pack(side=tk.LEFT, anchor=tk.N, fill=tk.BOTH, expand=tk.Y)
        self.ysbbb = ttk.Scrollbar(self.tag_frame, orient='vertical', command=self.tag_tree.yview)
        self.tag_tree.configure(yscrollcommand=self.ysbbb.set)
        self.ysbbb.pack(side=tk.RIGHT, anchor=tk.N, fill=tk.Y)
        self.tag_inner_frame.pack(side=tk.TOP, anchor=tk.N, fill=tk.BOTH, expand=True)

        # pack everything
        self.browsers.add(self.search_frame, text='search')
        self.browsers.add(self.tag_frame, text='tags')
        self.browsers.add(self.browse_frame, text='browse')

        self.frame.pack(fill=tk.BOTH, expand=True)
        self.browsers.pack(expand=True, fill=tk.BOTH)

        # finally populate the trees
        self.tree_reset()

    def search(self, *args):
        search_term = self.search_query.get()
        if search_term != "":
            res = self.db.search(search_term)  # n = 3 # for limiting
            self.search_tree.item("root", open=False)
            if self.search_tree.exists("search"):
                self.search_tree.delete("search")
            search_res = self.search_tree.insert('', 'end', iid="search", text='Search Results', open=True, values=['category'])
            for clip in res:
                self.search_tree.insert(search_res, 'end', text=clip.name, values=["clip", clip.f_name])
        else:
            if self.search_tree.exists("search"):
                self.search_tree.delete("search")
            self.search_tree.item("root", open=True)

    def tag_search(self, *args):
        search_term = self.tag_query.get()
        if search_term != "":
            self.tag_tree.item("root", open=False)
            if self.tag_tree.exists("search"):
                self.tag_tree.delete("search")
            search_res = self.tag_tree.insert('', 'end', iid="search", text='Search Results', open=True, values=['nop', 0])
            res_tags = self.db.tagdb.search(search_term)
            for tag, clips in res_tags:
                tag_folder = self.tag_tree.insert(search_res, 'end', text=tag, values=["tag", 0])
                for clip in clips:
                    self.tag_tree.insert(tag_folder, 'end', text=clip.name, values=["clip", clip.f_name])

        else:
            if self.tag_tree.exists("search"):
                self.tag_tree.delete("search")
            self.tag_tree.item("root", open=True)

    def last_search(self, *args):
        res = self.db.last_search
        if res is None:
            return
        self.search_tree.item("root", open=False)
        if self.search_tree.exists("search"):
            self.search_tree.delete("search")
        search_res = self.search_tree.insert('', 'end', iid="search", text='Search Results', open=True, values=['category'])
        for clip in res:
            self.search_tree.insert(search_res, 'end', text=clip.name, values=["clip", clip.f_name])

    def get_clip_from_event(self, event):
        if event.state != 8:  # sure numlock is on for 8 to work...
            if event.state != 0:
                return
        tv = event.widget
        if tv.identify_row(event.y) not in tv.selection():
            tv.selection_set(tv.identify_row(event.y))
        if not tv.selection():
            return
        item = tv.selection()[0]
        if tv.item(item, "values")[0] != 'clip':
            return
        clip_fname = tv.item(item, "values")[1]
        if clip_fname not in self.db.clips:
            return
        return self.db.clips[clip_fname]

    def make_drag_clip(self, event):
        the_clip = self.get_clip_from_event(event)
        if the_clip is None:
            return
        if dnd_start(DragClip(the_clip), event):
            pass

    def activate_clip_to_layer(self, event, layer):
        the_clip = self.get_clip_from_event(event)
        if the_clip is None:
            return
        self.backend.select_clip(the_clip, layer)

    def tree_reset(self):
        # search
        clips = self.db.alphabetical_listing
        if self.search_tree.exists("root"):
            self.search_tree.delete("root")
        self.s_tree_root = self.search_tree.insert('', 'end', iid="root", text='All', open=True, values=['category'])
        for c_name_clip in clips:
            self.search_tree.insert(self.s_tree_root, 'end', text=c_name_clip[0], values=["clip", c_name_clip[1].f_name])
        # browse
        files = self.db.hierarchical_listing
        for folder in self.folders.values():
            if self.browse_tree.exists(folder):
                self.browse_tree.delete(folder)
        if files is None:
            return
        self.folders = {}
        cur_folder = ''
        for i in range(len(files)):
            node = files[i]
            if node[0] == 'folder':
                top_folder = ''
                if node[2] in self.folders:
                    top_folder = self.folders[node[2]]
                    self.browse_tree.item(top_folder, open=True)
                self.folders[node[1]] = cur_folder = self.browse_tree.insert(top_folder, 'end', text=node[1], open=False, values=['folder'])
            else:
                self.browse_tree.insert(cur_folder, 'end', text=files[i][1], values=["clip", files[i][2]])
        # tags
        if self.tag_tree.exists("root"):
            self.tag_tree.delete("root")
        self.t_tree_root = self.tag_tree.insert('', 'end', iid="root", text='All', open=True, values=['nah', 0])

        all_tags = self.db.tagdb.search("")
        for tag, clips in all_tags:
            tag_folder = self.tag_tree.insert('root', 'end', text=tag, values=["tag", 0])
            for clip in clips:
                self.tag_tree.insert(tag_folder, 'end', text=clip.name, values=["clip", clip.f_name])


class ClipOrg:
    # popup with the clip org gui =)
    # ClipCont = ClipContainer but with unbind("<Double-3>")
    def __init__(self, root, parent):
        self.root = root
        self.parent = parent
        self.backend = self.parent.magi
        self.root.title('clip org')
        self.parent.root.call('wm', 'attributes', '.', '-topmost', '0')
        self.root.geometry('{}x{}'.format(C.THUMB_W * 4 + 100, 400))
        self.root.protocol("WM_DELETE_WINDOW", self.parent.exit_clip_org_gui)
        self.index = -1

        self.mainframe = tk.Frame(root)
        self.clip_frame = tk.Frame(self.mainframe)
        self.all_clip_frame = tk.Frame(self.clip_frame)
        self.all_clip_canvas = tk.Canvas(self.all_clip_frame)
        self.all_clip_inner_frame = tk.Frame(self.all_clip_canvas)
        self.vsb_all_clips = ttk.Scrollbar(self.all_clip_frame, orient="vertical", command=self.all_clip_canvas.yview)
        self.all_clip_canvas.configure(yscrollcommand=self.vsb_all_clips.set)

        self.search_query = tk.StringVar()
        self.search_field = tk.Entry(self.clip_frame, textvariable=self.search_query)
        self.search_query.trace('w', self.search)
        # self.search_field.bind('<Return>', self.search)

        self.clip_conts = {}
        self.clip_folds = []

        self.mainframe.pack(expand=True, fill="both")
        self.search_field.pack(side="top", fill="x")
        self.clip_frame.pack(side='top', expand=True, fill="both")
        self.all_clip_frame.pack(side='top', expand=True, fill="both")
        self.vsb_all_clips.pack(side="right", fill="y")
        self.all_clip_canvas.pack(side="left", fill="both", expand=True)
        self.all_clip_canvas.create_window((4, 4), window=self.all_clip_inner_frame, anchor="nw",
                                           tags="self.all_clip_inner_frame")

        self.all_clip_inner_frame.bind("<MouseWheel>", self.mouse_wheel)
        self.all_clip_inner_frame.bind("<Button-4>", self.mouse_wheel)
        self.all_clip_inner_frame.bind("<Button-5>", self.mouse_wheel)
        self.all_clip_inner_frame.bind("<Configure>", self.reset_scroll_region)

        self.initialize_all_clips()

    def reset_scroll_region(self, event=None):
        self.all_clip_canvas.configure(scrollregion=self.all_clip_canvas.bbox("all"))

    def mouse_wheel(self, event):
        self.all_clip_canvas.yview('scroll', -1 * int(event.delta // 120), 'units')

    def mouse_wheel_search(self, event):
        self.search_clip_canvas.yview('scroll', -1 * int(event.delta // 120), 'units')

    def initialize_all_clips(self):
        # first sort by filename : )
        all_clips = self.backend.db.alphabetical_listing  # [:50] speedup for testing..

        for i, cc in enumerate(all_clips):
            new_clip_cont = ClipOrgClip(self, self.all_clip_inner_frame, self.backend.select_clip, cc[1])
            r, c = (i // 4), (i % 4)
            new_clip_cont.grid(row=r, column=c)
            self.clip_conts[cc[1].f_name] = [new_clip_cont, r, c]

    def search(self, event, *args):
        # print('searching', self.search_query.get())
        search_term = self.search_query.get()
        self.search_res_clip_conts = []
        if search_term == "":
            for clip_cont in self.clip_conts.values():
                clip_cont[0].grid(row=clip_cont[1], column=clip_cont[2])
        else:
            res = self.backend.db.search(search_term)
            for clip_cont in self.clip_conts.values():
                clip_cont[0].grid_forget()
            # print(res)
            i = 0
            for clip in res:
                if clip.f_name in self.clip_conts:
                    self.clip_conts[clip.f_name][0].grid(row=i // 4, column=i % 4)
                    i += 1
        self.reset_scroll_region()

    def quit(self):
        self.backend.save_data()

    def close(self, *args):
        self.parent.root.call('wm', 'attributes', '.', '-topmost', str(int(self.parent.on_top_toggle.get())))
        self.root.destroy()


class ClipOrgClip(ClipContainer):
    def __init__(self, parent, parent_frame, selectfun, clip):
        self.selectfun = selectfun
        self.last_clip = None

        self.parent = parent
        self.backend = parent.backend
        self.root = parent.root
        self.frame = ttk.Frame(parent_frame, padding='5 0 5 0')
        self.frame = tk.Frame(parent_frame, padx=5, pady=0)

        self.grid = self.frame.grid
        self.grid_forget = self.frame.grid_forget

        self.imgs = []
        self.default_img = self.current_img = ImageTk.PhotoImage(Image.open(EMPTY_CLIP))
        self.current_img_i = 0
        self.hovered = False

        self.label = tk.Label(self.frame, image=self.current_img, text='test', compound='top', width=C.THUMB_W, bd=2)
        self.label.image = self.current_img
        self.label.pack()
        self.label.bind('<Double-1>', self.activate_l)
        self.label.bind('<Double-2>', self.regen_imgs)

        self.frame.dnd_accept = self.dnd_accept

        self.clip = None
        self.change_clip(clip)

        self.label.unbind('<Button-2>')
        for b in ["<MouseWheel>", "<Button-4>", "<Button-5>"]:
            self.label.bind(b, parent.mouse_wheel)
            self.frame.bind(b, parent.mouse_wheel)

    def change_clip(self, clip):
        if clip is None:
            self.remove_clip()
            return
        self.last_clip = self.clip
        self.clip = clip

        if self.clip.t_names is None:
            self.change_img_from_img(self.default_img)
        else:
            self.setup_imgs()
        self.toggle_dnd()

    def setup_imgs(self):
        f_names = [os.path.join(C.SCROT_DIR, t_name) for t_name in self.clip.t_names if os.path.exists(os.path.join(C.SCROT_DIR, t_name))]
        if len(f_names) > 0:
            f_name = f_names[0]
        else:
            f_name = EMPTY_CLIP
        self.imgs = [ImageTk.PhotoImage(Image.open(f_name))]
        if len(self.imgs) == 0:
            return
        self.current_img_i = 0
        self.change_img_from_img(self.imgs[self.current_img_i])

    def dnd_accept(self, source, event):
        pass

    def dnd_commit(self, source, event):
        pass



# this class will make thumbnails for you
import subprocess
import os

from PIL import Image, ImageChops, ImageOps

from sol.config import GlobalConfig
C = GlobalConfig()


class ThumbMaker:
    def __init__(self, desired_width, aspect_width=16, aspect_height=9):
        if not os.path.exists(C.SCROT_DIR):
            os.makedirs(C.SCROT_DIR)
        self.desired_width = desired_width
        self.desired_height = int(desired_width * aspect_height / aspect_width)

    @property
    def tempfile(self):
        return os.path.join(C.SCROT_DIR, 'temp.png')

    def correct_ff_path(self, cmd):
        return os.path.join(C.FFMPEG_PATH, cmd)

    def update_desired_width(self, new_width):
        self.desired_height = int(
            self.desired_height / self.desired_width * new_width)
        self.desired_width = new_width

    def quit(self):
        # clean up after yourself
        if os.path.exists(self.tempfile):
            os.remove(self.tempfile)

    def create_shot(self, f_in, f_out, seek_time=0, width=0, height=0):
        # makes a screenshot (used for when need to make a one-off thumbnail)
        # seek time is float from 0.0 to 1.0
        if width <= 0 or height <= 0:
            width, height = self.desired_width, self.desired_height
        if not os.path.exists(f_in):
            return
        if seek_time > 0:
            # want total length of video
            file_info = ''
            try:
                command = '{} -i "{}" -show_entries format=duration -v quiet -of csv="p=0"'.format(
                    self.correct_ff_path('ffprobe'), f_in)
                file_info = subprocess.check_output(command,
                                                    stderr=subprocess.STDOUT, shell=True)
            except subprocess.CalledProcessError as e:
                file_info = e.output
            file_info = float(file_info.decode('utf-8'))
            ss_param = seek_time * file_info
            command = '{0} -ss {1:.3f} -y -i "{2}" -q:v 2 -vframes 1 "{3}" -hide_banner -loglevel panic'\
                .format(self.correct_ff_path('ffmpeg'), ss_param, f_in, self.tempfile)
            print(command)
            print('#' * 40)
            process = subprocess.Popen(command)  # -ss to seek to frame
            process.communicate()
            if os.path.exists(self.tempfile):
                self.make_thumbnail(self.tempfile, f_out,
                                    new_size=(width, height))
                return f_out
            else:
                return
        else:
            command = '{0} -y -i "{1}" -vf "thumbnail" -q:v 2 -vframes 1 "{2}" -hide_banner -loglevel panic'\
                .format(self.correct_ff_path('ffmpeg'), f_in, self.tempfile)
            print(command)
            print('#' * 40)
            process = subprocess.Popen(command)  # -ss to seek to frame
            process.communicate()
            if os.path.exists(self.tempfile):
                self.make_thumbnail(self.tempfile, f_out,
                                    new_size=(width, height))
                return f_out
            else:
                return

    def nice_num(self, input_str):
        # takes a string and gives you a number to append
        # point of this is if you have the same filename in multiple folders
        # rather than overwriting the thumbnail we will (try) to make the filename
        # unique : _ )
        return sum([ord(c) for c in input_str])

    def gen_thumbnail(self, f_name, n_frames=1):
        # give a filename of a clip plus desired width and it will try to make a
        # thumbnail. alternatively tell it how many thumbnails you'd like and
        # the total number of frames and it will generate however many thumbnails
        # evenly dispersed
        if not os.path.exists(f_name):
            return

        # create output name
        split_f_name = os.path.split(os.path.splitext(f_name)[0])
        output_base_name = os.path.join(C.SCROT_DIR, split_f_name[1])
        bad_hash = self.nice_num(
            split_f_name[0]) * self.nice_num(split_f_name[1])
        while True:
            output_name = output_base_name + "_" + str(bad_hash) + "_0.png"
            if not os.path.exists(output_name):
                break
            bad_hash += 1

        if n_frames > 1:
            # list of all the thumbs
            tor = []
            # want total length of video
            file_info = ''
            try:
                command = '{} -i "{}" -show_entries format=duration -v quiet -of csv="p=0"'.format(
                    self.correct_ff_path('ffprobe'), f_name)
                file_info = subprocess.check_output(command,
                                                    stderr=subprocess.STDOUT, shell=True)
            except subprocess.CalledProcessError as e:
                file_info = e.output
            try:
                file_info = float(file_info.decode('utf-8'))
            except:
                print('your version of ffmpeg/ffprobe is probably too old')
                return tor

            # then get n frames spread out throughout the range of the entire clip
            interval = file_info / (n_frames + 1)
            for i in range(n_frames):
                seek_time = interval / 2 + i * interval
                command = '{0} -ss {1:.3f} -y -i "{2}" -vf "thumbnail" -q:v 2 -vframes 1 "{3}" -hide_banner -loglevel panic'\
                    .format(self.correct_ff_path('ffmpeg'), seek_time, f_name, self.tempfile)
                print(command)
                print('#' * 40)
                process = subprocess.Popen(command)  # -ss to seek to frame
                process.communicate()

                i_output_name = output_name[:-5] + '{}.png'.format(i)
                print(i_output_name)
                if os.path.exists(self.tempfile):
                    self.make_thumbnail(self.tempfile, i_output_name,
                                        new_size=(self.desired_width, self.desired_height))
                    i_output_name = os.path.split(i_output_name)[1]
                    tor += [i_output_name]
                else:
                    break
            return tor

        else:
            tor = self.create_shot(f_name, output_name)
            if tor is not None:
                return [tor]
            else:
                return

    def make_thumbnail(self, f_in, f_out, new_size=(100, 100), pad=True):
        # makes a thumbnail from a screenshot, pads it if the aspect ratio is off
        image = Image.open(f_in)
        image.thumbnail(new_size, Image.ANTIALIAS)
        image_size = image.size
        if pad:
            thumb = image.crop((0, 0, new_size[0], new_size[1]))
            offset_x = max((new_size[0] - image_size[0]) // 2, 0)
            offset_y = max((new_size[1] - image_size[1]) // 2, 0)
            thumb = ImageChops.offset(thumb, offset_x, offset_y)
        else:
            thumb = ImageOps.fit(image, new_size, Image.ANTIALIAS, (0.5, 0.5))
        thumb.save(f_out)



import tkinter as tk
from tkinter import ttk
import re
# import pprint
# pp = pprint.PrettyPrinter(indent=4)

from sol.config import GlobalConfig
C = GlobalConfig()


class MidiConfig:
    def __init__(self, root, parent):
        # needs to show up to the top right of the main window.
        self.root = root
        self.parent = parent
        self.midi_int = self.parent.magi.midi_interface

        self.selected_cmd = None
        self.hovered_cmd = None

        self.key_to_row = {}

        self.main_frame = ttk.Frame(self.root)
        self.main_frame.pack()

        self.midi_choice = tk.StringVar()

        self.input_label = ttk.Label(self.main_frame, text='input:')
        self.choose_midi = ttk.Combobox(self.main_frame, textvariable=self.midi_choice)
        self.refresh_inputs()
        self.choose_midi.config(state='readonly')
        self.choose_midi.set('None')

        self.midi_choice.trace("w", self.midi_choice_changed)

        self.refresh_inputs_but = ttk.Button(self.main_frame, text='[f5]', command=self.refresh_inputs, takefocus=False)
        # cmd name
        self.cmd_label = ttk.Label(self.main_frame, text='cmd:')
        self.cmd_text_var = tk.StringVar()
        self.cmd_text_label = ttk.Label(self.main_frame, textvariable=self.cmd_text_var)
        # cmd desc
        self.desc_label = ttk.Label(self.main_frame, text='desc:')
        self.desc_text_var = tk.StringVar()
        self.desc_text_label = ttk.Label(self.main_frame, textvariable=self.desc_text_var)
        # control type
        self.type_label = ttk.Label(self.main_frame, text='type:')
        self.key_type_choice = tk.StringVar()
        self.choose_type = ttk.Combobox(self.main_frame, textvariable=self.key_type_choice)
        type_options = ['----', 'togl', 'knob', 'sldr']
        self.choose_type.config(values=type_options)
        self.choose_type.config(state='readonly')
        self.choose_type.set('----')
        self.invert_type = tk.BooleanVar()
        self.invert_check_but = ttk.Checkbutton(self.main_frame, text='invert', variable=self.invert_type, takefocus=False)
        # scale factor
        self.factor_label = ttk.Label(self.main_frame, text='scale:')
        self.factor_var = tk.StringVar()
        self.factor_entry = tk.Spinbox(self.main_frame, from_=0, to=2, increment=0.005,
                                       textvariable=self.factor_var, justify=tk.CENTER, width=5)
        # configured keys
        self.keys_label = ttk.Label(self.main_frame, text='saved keys:')
        self.keys_text_var = tk.StringVar()
        self.keys_text_area = ttk.Label(self.main_frame, textvariable=self.keys_text_var)
        # save/clear buts
        self.save_but = ttk.Button(self.main_frame, text='save', command=self.save_cmd, takefocus=False)
        self.clear_but = ttk.Button(self.main_frame, text='clear', command=self.clear_cmd, takefocus=False)
        # gets filled in with midi keys
        self.key_fill_area = ttk.Frame(self.main_frame, height=6 * C.FONT_HEIGHT)
        self.key_fill_area.grid_propagate(False)
        ttk.Label(self.main_frame, text='key').grid(row=7, column=0, columnspan=2, sticky='we')
        ttk.Label(self.main_frame, text='values').grid(row=7, column=2, sticky='we')
        self.key_fill_row = ttk.Frame(self.key_fill_area)

        # grid it
        # input choice
        self.input_label.grid(row=0, column=0, sticky='we')
        self.choose_midi.grid(row=0, column=1, sticky='we')
        self.refresh_inputs_but.grid(row=0, column=2, sticky='we')
        # cmd info
        self.cmd_label.grid(row=1, column=0, sticky='we')
        self.cmd_text_label.grid(row=1, column=1, columnspan=2, sticky='we')
        self.desc_label.grid(row=2, column=0, sticky='we')
        self.desc_text_label.grid(row=2, column=1, columnspan=2, sticky='we')
        # configured keys
        self.keys_label.grid(row=3, column=0, sticky='we')
        self.keys_text_area.grid(row=3, column=1, columnspan=2, sticky='we')
        # key type
        self.type_label.grid(row=4, column=0, sticky='we')
        self.choose_type.grid(row=4, column=1, sticky='we')
        self.invert_check_but.grid(row=4, column=2, sticky='we')
        # scale factor
        self.factor_label.grid(row=5, column=0, columnspan=2, sticky='we')
        self.factor_entry.grid(row=5, column=2, sticky='we')
        # save/clear buts
        self.save_but.grid(row=6, column=0, sticky='we')
        self.clear_but.grid(row=6, column=2, sticky='we')
        self.key_fill_area.grid(row=8, rowspan=5, column=0, columnspan=3, sticky='nwes')
        self.key_fill_row.grid(row=0, rowspan=5, column=0, columnspan=3, sticky='news')

        self.overlay = MidiOverlay(self, self.parent)
        self.root.protocol("WM_DELETE_WINDOW", self.close)
        self.root.title('config midi')
        # undo topmost
        self.parent.root.call('wm', 'attributes', '.', '-topmost', '0')

        self.midi_int.enter_config_mode(self)

    # midi interface related

    def refresh_inputs(self, *args):
        self.midi_int.refresh_input_ports()
        input_options = ['None'] + list(self.midi_int.input_ports.keys()) + ['osc']
        self.choose_midi.config(values=input_options)
        self.choose_midi.set('None')

    def midi_choice_changed(self, *args):
        new_choice = self.cur_input_name
        self.midi_int.close_all_inputs()
        if new_choice != 'None' and new_choice != 'osc':
            self.midi_int.open_midi_input(new_choice)
        all_ks = list(self.key_to_row.keys())
        for k in all_ks:
            self.drop_key(k)
        self.overlay.recolor_everything()

    def save_cmd(self, *args):
        sc = self.selected_cmd
        cur_inp = self.cur_input_name
        if sc is None or cur_inp is None:
            return

        # if haven't selected any new keys then don't override
        pos_keys = self.get_chosen_keys()
        if len(pos_keys) == 0:
            if cur_inp in self.midi_int.name_to_cmd[sc]['midi_keys']:
                pos_keys = self.midi_int.name_to_cmd[sc]['midi_keys'][cur_inp]['keys']

        factor = self.factor_var.get()
        try:
            factor = float(factor)
        except:
            factor = None

        self.midi_int.name_to_cmd[sc]['midi_keys'][cur_inp] = {
            'control': self.cur_input_name,
            'keys': pos_keys,
            'type': self.key_type_choice.get(),
            'invert': self.invert_type.get(),
            'factor': factor
        }

        self.keys_text_var.set(', '.join(pos_keys))

        # pp.pprint(self.midi_int.name_to_cmd[sc]['midi_keys'])

    def clear_cmd(self, *args):
        sc = self.selected_cmd
        cur_inp = self.cur_input_name
        if sc is None:
            return
        if cur_inp in self.midi_int.name_to_cmd[sc]['midi_keys']:
            del self.midi_int.name_to_cmd[sc]['midi_keys'][cur_inp]
        self.choose_type.set('----')
        self.invert_type.set(False)
        self.keys_text_var.set('')
        self.unselect_all()
        self.reset_sens_val()

    # key fill

    def add_key(self, key):
        key_row = ttk.Frame(self.key_fill_row)
        key_label = ttk.Label(key_row, text=key, width=10)
        key_sel_var = tk.BooleanVar()
        key_checkbox = ttk.Checkbutton(key_row, variable=key_sel_var, takefocus=False)
        key_vals_var = tk.StringVar()
        key_vals_label = ttk.Label(key_row, textvariable=key_vals_var)
        key_label.pack(side=tk.LEFT)
        key_checkbox.pack(side=tk.LEFT)
        key_vals_label.pack(side=tk.LEFT, fill=tk.X, expand=True)
        key_row.pack(side=tk.TOP, anchor='w', expand=True, fill=tk.X)

        self.key_to_row[key] = [key_row, key_sel_var, key_vals_var]

    def drop_key(self, key):
        if key in self.key_to_row:
            self.key_to_row[key][0].pack_forget()
            del self.key_to_row[key]

    def update_vals(self, *args):
        updates = self.midi_int.midi_classify()
        if updates is not None:
            for k, v in updates:
                self.update_key_vals(k, v)

    def update_key_vals(self, key, vals):
        if key in self.key_to_row:
            self.key_to_row[key][-1].set(str(vals))

    def get_chosen_keys(self):
        tor = []
        for k, vs in self.key_to_row.items():
            if vs[1].get():
                tor.append(k)
        return tor

    def unselect_all(self):
        for k, vs in self.key_to_row.items():
            vs[1].set(False)

    # overlay related

    def hovered(self):
        hc = self.hovered_cmd
        if hc is not None and hc in self.midi_int.name_to_cmd:
            self.cmd_text_var.set(hc)
            self.desc_text_var.set(self.midi_int.name_to_cmd[hc]['desc'])

    def selected(self):
        sc = self.selected_cmd
        if sc is not None and sc in self.midi_int.name_to_cmd:
            saved = self.midi_int.name_to_cmd[sc]['midi_keys']
            cur_inp = self.cur_input_name
            self.unselect_all()
            self.reset_sens_val()
            if cur_inp is not None and cur_inp in saved:
                k_type, inv = saved[cur_inp]['type'], saved[cur_inp]['invert']
                keys = saved[cur_inp]['keys']
                key_text = ', '.join(keys)
                self.choose_type.set(k_type)
                self.invert_type.set(inv)
                self.keys_text_var.set(key_text)
                return

        self.choose_type.set('----')
        self.invert_type.set(False)
        self.keys_text_var.set('')

    def unhovered(self):
        sc = self.selected_cmd
        if sc is not None and sc in self.midi_int.name_to_cmd:
            self.cmd_text_var.set(sc)
            self.desc_text_var.set(self.midi_int.name_to_cmd[sc]['desc'])
        else:
            self.cmd_text_var.set('')
            self.desc_text_var.set('')

    # helpers

    @property
    def cur_input_name(self):
        inp = self.midi_choice.get()
        if inp != 'None':
            return inp

    def reset_sens_val(self):
        sc = self.selected_cmd

        # default
        self.factor_var.set('----')
        self.factor_entry.config(state='disabled')

        if sc is None:
            return

        sens_factor = self.midi_int.name_to_cmd[sc]['factor']
        # if it has a factor
        if sens_factor is not None:
            cur_inp = self.cur_input_name
            saved = self.midi_int.name_to_cmd[sc]['midi_keys']
            # check if it has a saved factor value
            if cur_inp is not None and cur_inp in saved:
                saved_sens_factor = saved[cur_inp]['factor']
                if saved_sens_factor is not None:
                    sens_factor[0] = saved_sens_factor
            if not sens_factor[1]:
                self.factor_entry.config(state='normal')
            self.factor_var.set(sens_factor[0])

    def close(self):
        if self.overlay is not None:
            self.overlay.close()
        self.root.destroy()
        # pp.pprint(self.midi_int.gen_savedata())
        self.parent.magi.save_midi()
        self.parent.midi_config_gui = None
        # redo topmost setting
        self.parent.root.call('wm', 'attributes', '.', '-topmost', str(int(self.parent.on_top_toggle.get())))


class MidiOverlay:
    def __init__(self, parent, base_gui):
        self.parent = parent
        self.base_gui = base_gui
        self.last_pos = ""
        self.offsets = [0, 0]
        self.spacing = 1
        self.base_coords = [0, 0]
        self.geo_regex = re.compile(r'(\d+)x(\d+)\+(\-?\d+)\+(-?\d+)')

        self.wname_to_widget = {}
        self.wname_to_rect = {}

        self.root = tk.Toplevel()
        self.canvas = tk.Canvas(self.root, bg="black")
        self.canvas.pack()

        self.root.wm_attributes("-topmost", True)
        self.root.wm_attributes("-transparentcolor", "black")
        self.root.attributes('-alpha', 0.5)
        self.root.overrideredirect(True)

        self.recursive_buildup()

        self.base_gui.root.bind('<Configure>', self.base_moved)

        self.base_moved()

    # helpers

    def recursive_buildup(self, start_el=None):
        if start_el is None:
            start_el = self.base_gui.root
        for c in start_el.winfo_children():
            if c.winfo_name() in self.parent.midi_int.all_wnames:
                self.wname_to_widget[c.winfo_name()] = c
                # print('found', c.winfo_name())
            self.recursive_buildup(c)

    def close(self):
        self.base_gui.root.unbind('<Configure>')
        self.root.destroy()

    # draw funs

    def draw_everything(self):
        self.canvas.delete("all")
        for k in self.wname_to_widget:
            # need to do multiple items
            if k in self.parent.midi_int.wname_to_names:
                self.draw_multiple_widgets(self.parent.midi_int.wname_to_names[k], k)
            else:
                self.draw_single_widget(self.wname_to_widget[k], k)
        self.canvas.addtag_all('overlay')
        self.create_binds()

    def recolor_everything(self):
        for k, r in self.wname_to_rect.items():
            if k != self.parent.selected_cmd:
                self.canvas.itemconfig(r, fill=self.get_widget_color(k))

    def get_widget_color(self, w_name):
        color_get = 'empty'
        if w_name is not None and w_name in self.parent.midi_int.name_to_cmd:
            if self.parent.cur_input_name in self.parent.midi_int.name_to_cmd[w_name]['midi_keys']:
                color_get = 'set'
        return C.CURRENT_THEME.midi_setting_colors[color_get]

    def draw_widget(self, coords, w_name, s=0):
        [x, y, w, h] = coords
        if w_name in self.parent.midi_int.double_width_pls:
            w *= 2
        x1, y1 = x - self.base_coords[0] + s, y - self.base_coords[1] + s
        x2, y2 = x1 + w - 2 * s, y1 + h - 2 * s
        # check fill
        self.wname_to_rect[w_name] = self.canvas.create_rectangle(x1, y1, x2, y2, tag=w_name,
                                                                  fill=self.get_widget_color(w_name))

    def draw_single_widget(self, w, w_name):
        # give a tag to fake square that corresponds to midi's cmd_name
        x, y, width, height = w.winfo_rootx(), w.winfo_rooty(), w.winfo_width(), w.winfo_height()
        self.draw_widget([x, y, width, height], w_name)

    def draw_multiple_widgets(self, ws, w_key):
        top_w = self.wname_to_widget[w_key]
        bx, by, tw, th = top_w.winfo_rootx(), top_w.winfo_rooty(), top_w.winfo_width(), top_w.winfo_height()
        # determine how to split, i only deal with 2 or 4..
        n = len(ws)
        if tw > th:
            if n > 3:
                # sideways then vertical
                tot_c = n // 2
                w, h = tw // 2, th // tot_c
                for i, wn in enumerate(ws):
                    r, c = i // 2, i % 2
                    x, y = bx + (c * w), by + (r * h)
                    self.draw_widget([x, y, w, h], wn, self.spacing)
            else:
                # horizontal all the way
                w = tw / n
                for i, wn in enumerate(ws):
                    x = bx + (i * w)
                    self.draw_widget([x, by, w, th], wn, self.spacing)
        else:
            # vertical all the way
            h = th / n
            for i, wn in enumerate(ws):
                y = by + (i * h)
                self.draw_widget([bx, y, tw, h], wn, self.spacing)

    # events bindings

    def create_binds(self):
        self.canvas.tag_bind("overlay", "<Enter>", self.hover_check)
        self.canvas.tag_bind("overlay", "<Leave>", self.hover_restore)
        self.canvas.bind("<ButtonPress-1>", self.select_bind)

    def base_moved(self, *args):
        base_pos = self.base_gui.root.geometry()
        if self.last_pos != base_pos:
            new_dims = list(map(int, self.geo_regex.match(base_pos).groups()))
            self.base_coords = new_dims[2:]
            # index of first plus, before this is width/height
            wi = base_pos.index('+')
            if base_pos[:wi] != self.last_pos[:wi]:
                self.offsets = [self.base_gui.root.winfo_rootx() - new_dims[2],
                                self.base_gui.root.winfo_rooty() - new_dims[3]]
                self.canvas.configure(width=new_dims[0] + self.offsets[0],
                                      height=new_dims[1] + self.offsets[1])
                # redraw everything..
                self.draw_everything()
            self.root.geometry('+{}+{}'.format(*new_dims[2:]))
            self.parent.root.geometry('+{}+{}'.format(self.base_gui.root.winfo_rootx() + new_dims[0],
                                                      self.base_gui.root.winfo_rooty() - self.offsets[1]))
            self.last_pos = base_pos

    def hover_check(self, event):
        item = self.canvas.find_closest(self.canvas.canvasx(event.x), self.canvas.canvasy(event.y), halo=5)[0]
        self.parent.hovered_cmd = self.canvas.gettags(item)[0]
        self.parent.hovered()

    def hover_restore(self, event):
        self.parent.hovered_cmd = None
        self.parent.unhovered()

    def select_bind(self, event):
        scmd = self.parent.selected_cmd
        if scmd in self.wname_to_rect:
            # restore color
            self.canvas.itemconfig(self.wname_to_rect[scmd],
                                   fill=self.get_widget_color(scmd))
        item = self.canvas.find_closest(self.canvas.canvasx(event.x), self.canvas.canvasy(event.y), halo=5)[0]
        self.parent.selected_cmd = self.canvas.gettags(item)[0]
        self.parent.selected()
        self.canvas.itemconfig(self.wname_to_rect[self.parent.selected_cmd],
                               fill=C.CURRENT_THEME.midi_setting_colors['selected'])


if __name__ == '__main__':

    rootwin = tk.Tk()
    rootwin.title('midi config')

    test_config = MidiConfig(rootwin)
    rootwin.mainloop()



default_style = 'clam'

default_font = None

# {
#     'size': 9,
#     'weight': 'normal',
#     'underline': 0,
#     'overstrike': 0,
#     'slant': 'roman',
#     # 'family': 'Segoe UI'
#     'family': 'Tamsyn6x12'
# }

default_colors = {
    'foreground': 'white',
    'activeForeground': '#222',
    'disabledForeground': '#000',
    'background': '#131313',
    # 'activeBackground': '', # menus
    'insertBackground': '#367AE3',
    'selectColor': '#367AE3',
    # these are for selecting text w/in entries?
    # but not ttk...
    'selectForeground': '#222',
    'selectBackground': '#367AE3',
    'troughColor': '#232626',
    # 'highlightBackground': '',  # this goes around canvases for some reason
    # 'highlightColor': '',
}

# for consistency
default_colors['highlightBackground'] = default_colors['background']

# http://paletton.com/#uid=7001G0k72lzjLdFijmMbwigdZeD
pad_colors = [
   # primary color
   ['#AC8686',
    '#6D2A2A'],
   # secondary color 1
   ['#ACA986',
    '#6D682A'],
   # secondary color 2
   ['#675D74',
    '#33204A'],
   # complement color
   ['#6B896B',
    '#215721']
]

delete_bg = '#661A1A'

midi_setting_colors = {
  'selected': '#345486',
  'empty': '#F0F0F0',
  'set': '#6BB321',
}

# ttk styled specific colors
styled_button_color = '#2B2B2B'
styled_button_border = '#1E1E1E'
styled_button_disabled_color = '#202020'
styled_button_disabled_color_darker = '#17191D'

# list of tuples of style to replace & a dict of all their options
ttk_style_opts_to_config = [ 
('TButton', {
  'background': styled_button_color,
  'bordercolor': styled_button_border,
  'lightcolor': styled_button_border,
  'darkcolor': styled_button_border,
}),
('TLabel', {
  'background': styled_button_color,
  'bordercolor': styled_button_border,
  'lightcolor': styled_button_border,
  'darkcolor': styled_button_border,
}),
('TFrame', {
  'bordercolor': styled_button_color,
  'lightcolor': styled_button_border,
  'darkcolor': styled_button_border,
}),
('fakesep.TFrame', {
  'bordercolor': styled_button_color,
  'lightcolor': styled_button_color,
  'darkcolor': styled_button_color,
  'background': styled_button_border,
}),
('TEntry', {
  'fieldbackground': styled_button_color,
  'background': styled_button_color,
  'selectbackground': default_colors['selectColor'],
  'bordercolor': styled_button_color,
  'lightcolor': styled_button_color,
  'darkcolor': styled_button_color,
}),
('TCheckbutton', {
  'background': default_colors['background'],
}),
('TRadiobutton', {
  'background': default_colors['background'],
}),
('Treeview', {
  'indent': 2,
  'borderwidth':  0,
  'relief': 'flat',
  'background': default_colors['activeForeground'],
  'fieldbackground': default_colors['activeForeground'],
  'selectforeground': default_colors['selectColor'],
  'bordercolor': styled_button_color,
  'lightcolor': styled_button_color,
  'darkcolor': styled_button_color,
}),
('Treeview.Heading', {
  'background': default_colors['background'],
}),
('TNotebook', {
  'relief': 'flat',
  'background': default_colors['background'],
  'bordercolor': styled_button_color,
  'lightcolor': styled_button_color,
  'darkcolor': styled_button_color,
}),
('TNotebook.Tab', {
  'relief': 'flat',
  'background': styled_button_disabled_color,
  'bordercolor': styled_button_disabled_color,
  'lightcolor': styled_button_color,
  'darkcolor': styled_button_color,
}),
('TLabelFrame', {
  'relief': 'flat',
  'background': default_colors['background'],
  'bordercolor': styled_button_color,
  'lightcolor': styled_button_color,
  'darkcolor': styled_button_color,
}),
('Vertical.TScrollbar', {
  'background': styled_button_color,
  'bordercolor': styled_button_color,
  'troughcolor': styled_button_disabled_color_darker,
  'lightcolor': styled_button_disabled_color,
  'darkcolor': styled_button_disabled_color_darker,
}),
('Horizontal.TScrollbar', {
  'background': styled_button_color,
  'bordercolor': styled_button_color,
  'troughcolor': styled_button_disabled_color_darker,
  'lightcolor': styled_button_disabled_color,
  'darkcolor': styled_button_disabled_color_darker,
}),
('Vertical.TScale', {
  'background': styled_button_color,
  'bordercolor': styled_button_color,
  'troughcolor': styled_button_disabled_color_darker,
  'lightcolor': styled_button_disabled_color,
  'darkcolor': styled_button_disabled_color_darker,
}),
('Horizontal.TScale', {
  'background': styled_button_color,
  'bordercolor': styled_button_color,
  'troughcolor': styled_button_disabled_color_darker,
  'lightcolor': styled_button_disabled_color,
  'darkcolor': styled_button_disabled_color_darker,
}),
('Combobox.field', {
  'background': styled_button_color,
  'fieldbackground': styled_button_color,
  'bordercolor': styled_button_color,
  'lightcolor': styled_button_color,
  'darkcolor': styled_button_color,
}),

]

# style options that need to be mapped (for active/disabled states usually)
ttk_style_opts_to_map = [
('TButton', {
    'background': [('disabled', styled_button_disabled_color), ('active', styled_button_disabled_color_darker)],
    'relief': [('pressed', '!disabled', 'sunken')]
}),
('TLabel', {
    'background': [('disabled', styled_button_disabled_color), ('active', styled_button_color)],
}),
('fakebut.TLabel', {
    'background': [('disabled', styled_button_disabled_color_darker), ('active', styled_button_color)],
    'foreground': [('disabled', default_colors['foreground']), ('active', default_colors['foreground'])],
}),
('TCheckbutton', {
    'background': [('disabled', styled_button_disabled_color), ('active', styled_button_color)],
}),
('TRadiobutton', {
    'background': [('disabled', styled_button_disabled_color), ('active', styled_button_color)],
}),
('Treeview', {
   'background': [('selected', 'focus', default_colors['selectColor']), ('selected', '!focus', '#6087C4')],
}),
('Treeview.Heading', {
   'background': [('selected', 'focus', default_colors['selectColor']), ('selected', '!focus', '#6087C4')],
}),
('TNotebook.Tab', {
   'background': [('selected', styled_button_color)],
}),
('Vertical.TScrollbar', {
    'background': [('disabled', styled_button_disabled_color), ('active', default_colors['troughColor'])],
}),
('Horizontal.TScrollbar', {
    'background': [('disabled', styled_button_disabled_color), ('active', default_colors['troughColor'])],
}),
('Vertical.TScale', {
    'background': [('disabled', styled_button_disabled_color), ('active', default_colors['troughColor'])],
}),
('Horizontal.TScale', {
    'background': [('disabled', styled_button_disabled_color), ('active', default_colors['troughColor'])],
}),
('TCombobox', {
    'fieldbackground': [('readonly', 'focus', '#6087C4'), ('readonly', styled_button_disabled_color)],
    'background': [('active', styled_button_color), ('pressed', styled_button_border)],
    'foreground': [('readonly', 'focus', default_colors['selectForeground'])],
}),
]



class IsadoraBL:
    """
    defines the isadorabl api
    """
    def __init__(self,no_layers):
        self.no_layers = 2 #no_layers

        # for updating current clip positions
        self.current_clip_pos = [None] * no_layers
        # index current_clip_pos by layer
        self.clip_pos_addr = [None] * no_layers
        # keep track of last speed
        self.current_clip_spd = [1] * no_layers

        for n in range(no_layers):
            self.clip_pos_addr[n] = "/layer{}/position".format(n+1)

        self.external_looping = True
        self.lt_lu = {'d':1,'b':2} # loop type lookup


    def play(self,layer):
        last_spd = self.current_clip_spd[layer]
        if last_spd < 0:
            last_spd *= -1
        return self.set_playback_speed(layer,last_spd)

    def pause(self,layer):
        last_spd = self.current_clip_spd[layer]
        tor = self.set_playback_speed(layer,0)      
        self.current_clip_spd[layer] = last_spd
        return tor

    def reverse(self,layer):
        last_spd = self.current_clip_spd[layer]
        if last_spd > 0:
            last_spd *= -1
        return self.set_playback_speed(layer,last_spd)

    # not implemented yet
    # have to import random
    # then call random.random()
    def random(self,layer):
        return ('/random',1)

    def set_clip_pos(self,layer,pos):
        addr = "/layer{}/seek".format(layer+1)
        return (addr, pos*100)

    def select_clip(self,layer,clip):
        addr = "/layer{}/clip".format(layer+1)
        return (addr, int(clip.command))

    def clear_clip(self,layer):
        addr = "/layer{}/clip".format(layer+1)
        return (addr, 0)

    def set_playback_speed(self,layer,speed):
        self.current_clip_spd[layer] = speed
        addr = "/layer{}/speed".format(layer+1)
        return (addr, speed) 

    ### looping
    def set_loop_type(self,layer,lt):
        addr = "/layer{}/loop".format(layer+1)
        loop_type = self.lt_lu[lt]
        return (addr, loop_type)

    def set_loop_a(self,layer,clip,a=None,b=None):
        if a is None: a = 0
        addr = "/layer{}/loop/a".format(layer+1)
        a = 100 * a
        return (addr, a)

    def set_loop_b(self,layer,clip,a=None,b=None):
        if a is None: a = 0
        if b is None: b = 1
        addr = "/layer{}/loop/b".format(layer+1)
        b = 100 * (b-a) # actually duration
        return (addr, b)
